# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import numpy as np
import pytest

import awkward as ak


def test_example():
    layout, report = ak.typetracer.typetracer_with_report(
        {
            "class": "ListOffsetArray",
            "offsets": "i64",
            "content": {
                "class": "NumpyArray",
                "primitive": "float64",
                "form_key": "node1",
            },
            "form_key": "node0",
        }
    )
    assert len(report.shape_touched) == 0
    tmp = layout.form  # noqa: F841
    assert len(report.shape_touched) == 0


def test_NumpyArray():
    v2a = ak.contents.numpyarray.NumpyArray(
        np.array([0.0, 1.1, 2.2, 3.3]),
        parameters={"some": "stuff", "other": [1, 2, "three"]},
    )

    layout, report = ak.typetracer.typetracer_with_report(v2a.form_with_key())
    assert len(report.shape_touched) == 0
    tmp = layout.form  # noqa: F841
    assert len(report.shape_touched) == 0


def test_EmptyArray():
    v2a = ak.contents.emptyarray.EmptyArray()

    layout, report = ak.typetracer.typetracer_with_report(v2a.form_with_key())
    assert len(report.shape_touched) == 0
    tmp = layout.form  # noqa: F841
    assert len(report.shape_touched) == 0


def test_NumpyArray_shape():
    v2a = ak.contents.numpyarray.NumpyArray(
        np.arange(2 * 3 * 5, dtype=np.int64).reshape(2, 3, 5)
    )

    layout, report = ak.typetracer.typetracer_with_report(v2a.form_with_key())
    assert len(report.shape_touched) == 0
    tmp = layout.form  # noqa: F841
    assert len(report.shape_touched) == 1  # this one actually *should* touch


@pytest.mark.parametrize("flatlist_as_rvec", [False, True])
def test_RegularArray_NumpyArray(flatlist_as_rvec):
    v2a = ak.contents.regulararray.RegularArray(
        ak.contents.numpyarray.NumpyArray(np.array([0.0, 1.1, 2.2, 3.3, 4.4, 5.5])),
        3,
    )

    layout, report = ak.typetracer.typetracer_with_report(v2a.form_with_key())
    assert len(report.shape_touched) == 0
    tmp = layout.form  # noqa: F841
    assert len(report.shape_touched) == 0


def test_RegularArray_NumpyArray_v2b():
    v2b = ak.contents.regulararray.RegularArray(
        ak.contents.emptyarray.EmptyArray().to_NumpyArray(np.dtype(np.float64)),
        0,
        zeros_length=10,
    )

    layout, report = ak.typetracer.typetracer_with_report(v2b.form_with_key())
    assert len(report.shape_touched) == 0
    tmp = layout.form  # noqa: F841
    assert len(report.shape_touched) == 0


@pytest.mark.parametrize("flatlist_as_rvec", [False, True])
def test_ListArray_NumpyArray(flatlist_as_rvec):
    v2a = ak.contents.listarray.ListArray(
        ak.index.Index(np.array([4, 100, 1], np.int64)),
        ak.index.Index(np.array([7, 100, 3, 200], np.int64)),
        ak.contents.numpyarray.NumpyArray(
            np.array([6.6, 4.4, 5.5, 7.7, 1.1, 2.2, 3.3, 8.8])
        ),
    )

    layout, report = ak.typetracer.typetracer_with_report(v2a.form_with_key())
    assert len(report.shape_touched) == 0
    tmp = layout.form  # noqa: F841
    assert len(report.shape_touched) == 0


@pytest.mark.parametrize("flatlist_as_rvec", [False, True])
def test_ListOffsetArray_NumpyArray(flatlist_as_rvec):
    v2a = ak.contents.listoffsetarray.ListOffsetArray(
        ak.index.Index(np.array([1, 4, 4, 6, 7], np.int64)),
        ak.contents.numpyarray.NumpyArray(
            np.array([6.6, 1.1, 2.2, 3.3, 4.4, 5.5, 7.7])
        ),
    )

    layout, report = ak.typetracer.typetracer_with_report(v2a.form_with_key())
    assert len(report.shape_touched) == 0
    tmp = layout.form  # noqa: F841
    assert len(report.shape_touched) == 0


def test_RecordArray_NumpyArray():
    v2a = ak.contents.recordarray.RecordArray(
        [
            ak.contents.numpyarray.NumpyArray(np.array([0, 1, 2, 3, 4], np.int64)),
            ak.contents.numpyarray.NumpyArray(np.array([0.0, 1.1, 2.2, 3.3, 4.4, 5.5])),
        ],
        ["x", "y"],
        parameters={"__record__": "Something"},
    )

    layout, report = ak.typetracer.typetracer_with_report(v2a.form_with_key())
    assert len(report.shape_touched) == 0
    tmp = layout.form
    assert len(report.shape_touched) == 0

    v2b = ak.contents.recordarray.RecordArray(
        [
            ak.contents.numpyarray.NumpyArray(np.array([0, 1, 2, 3, 4], np.int64)),
            ak.contents.numpyarray.NumpyArray(np.array([0.0, 1.1, 2.2, 3.3, 4.4, 5.5])),
        ],
        None,
    )

    layout, report = ak.typetracer.typetracer_with_report(v2b.form_with_key())
    assert len(report.shape_touched) == 0
    tmp = layout.form
    assert len(report.shape_touched) == 0

    v2c = ak.contents.recordarray.RecordArray([], [], 10)

    layout, report = ak.typetracer.typetracer_with_report(v2c.form_with_key())
    assert len(report.shape_touched) == 0
    tmp = layout.form
    assert len(report.shape_touched) == 0

    v2d = ak.contents.recordarray.RecordArray([], None, 10)

    layout, report = ak.typetracer.typetracer_with_report(v2d.form_with_key())
    assert len(report.shape_touched) == 0
    tmp = layout.form  # noqa: F841
    assert len(report.shape_touched) == 0


def test_IndexedArray_NumpyArray():
    v2a = ak.contents.indexedarray.IndexedArray(
        ak.index.Index(np.array([2, 2, 0, 1, 4, 5, 4], np.int64)),
        ak.contents.numpyarray.NumpyArray(np.array([0.0, 1.1, 2.2, 3.3, 4.4, 5.5])),
    )

    layout, report = ak.typetracer.typetracer_with_report(v2a.form_with_key())
    assert len(report.shape_touched) == 0
    tmp = layout.form  # noqa: F841
    assert len(report.shape_touched) == 0


def test_IndexedOptionArray_NumpyArray():
    v2a = ak.contents.indexedoptionarray.IndexedOptionArray(
        ak.index.Index(np.array([2, 2, -1, 1, -1, 5, 4], np.int64)),
        ak.contents.numpyarray.NumpyArray(np.array([0.0, 1.1, 2.2, 3.3, 4.4, 5.5])),
    )

    layout, report = ak.typetracer.typetracer_with_report(v2a.form_with_key())
    assert len(report.shape_touched) == 0
    tmp = layout.form  # noqa: F841
    assert len(report.shape_touched) == 0


def test_ByteMaskedArray_NumpyArray():
    v2a = ak.contents.bytemaskedarray.ByteMaskedArray(
        ak.index.Index(np.array([1, 0, 1, 0, 1], np.int8)),
        ak.contents.numpyarray.NumpyArray(np.array([1.1, 2.2, 3.3, 4.4, 5.5, 6.6])),
        valid_when=True,
    )

    layout, report = ak.typetracer.typetracer_with_report(v2a.form_with_key())
    assert len(report.shape_touched) == 0
    tmp = layout.form
    assert len(report.shape_touched) == 0

    v2b = ak.contents.bytemaskedarray.ByteMaskedArray(
        ak.index.Index(np.array([0, 1, 0, 1, 0], np.int8)),
        ak.contents.numpyarray.NumpyArray(np.array([1.1, 2.2, 3.3, 4.4, 5.5, 6.6])),
        valid_when=False,
    )

    layout, report = ak.typetracer.typetracer_with_report(v2b.form_with_key())
    assert len(report.shape_touched) == 0
    tmp = layout.form  # noqa: F841
    assert len(report.shape_touched) == 0


def test_BitMaskedArray_NumpyArray():
    v2a = ak.contents.bitmaskedarray.BitMaskedArray(
        ak.index.Index(
            np.packbits(
                np.array(
                    [
                        1,
                        1,
                        1,
                        1,
                        0,
                        0,
                        0,
                        0,
                        1,
                        0,
                        1,
                        0,
                        1,
                    ],
                    np.uint8,
                )
            )
        ),
        ak.contents.numpyarray.NumpyArray(
            np.array(
                [0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6]
            )
        ),
        valid_when=True,
        length=13,
        lsb_order=False,
    )

    layout, report = ak.typetracer.typetracer_with_report(v2a.form_with_key())
    assert len(report.shape_touched) == 0
    tmp = layout.form
    assert len(report.shape_touched) == 0

    v2b = ak.contents.bitmaskedarray.BitMaskedArray(
        ak.index.Index(
            np.packbits(
                np.array(
                    [
                        0,
                        0,
                        0,
                        0,
                        1,
                        1,
                        1,
                        1,
                        0,
                        1,
                        0,
                        1,
                        0,
                    ],
                    np.uint8,
                )
            )
        ),
        ak.contents.numpyarray.NumpyArray(
            np.array(
                [0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6]
            )
        ),
        valid_when=False,
        length=13,
        lsb_order=False,
    )

    layout, report = ak.typetracer.typetracer_with_report(v2b.form_with_key())
    assert len(report.shape_touched) == 0
    tmp = layout.form
    assert len(report.shape_touched) == 0

    v2c = ak.contents.bitmaskedarray.BitMaskedArray(
        ak.index.Index(
            np.packbits(
                np.array(
                    [
                        0,
                        0,
                        0,
                        0,
                        1,
                        1,
                        1,
                        1,
                        0,
                        0,
                        0,
                        1,
                        0,
                        1,
                        0,
                        1,
                    ],
                    np.uint8,
                )
            )
        ),
        ak.contents.numpyarray.NumpyArray(
            np.array(
                [0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6]
            )
        ),
        valid_when=True,
        length=13,
        lsb_order=True,
    )

    layout, report = ak.typetracer.typetracer_with_report(v2c.form_with_key())
    assert len(report.shape_touched) == 0
    tmp = layout.form
    assert len(report.shape_touched) == 0

    v2d = ak.contents.bitmaskedarray.BitMaskedArray(
        ak.index.Index(
            np.packbits(
                np.array(
                    [
                        1,
                        1,
                        1,
                        1,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        1,
                        0,
                        1,
                        0,
                    ],
                    np.uint8,
                )
            )
        ),
        ak.contents.numpyarray.NumpyArray(
            np.array(
                [0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6]
            )
        ),
        valid_when=False,
        length=13,
        lsb_order=True,
    )

    layout, report = ak.typetracer.typetracer_with_report(v2d.form_with_key())
    assert len(report.shape_touched) == 0
    tmp = layout.form  # noqa: F841
    assert len(report.shape_touched) == 0


def test_UnmaskedArray_NumpyArray():
    v2a = ak.contents.unmaskedarray.UnmaskedArray(
        ak.contents.numpyarray.NumpyArray(np.array([0.0, 1.1, 2.2, 3.3]))
    )

    layout, report = ak.typetracer.typetracer_with_report(v2a.form_with_key())
    assert len(report.shape_touched) == 0
    tmp = layout.form  # noqa: F841
    assert len(report.shape_touched) == 0


def test_UnionArray_NumpyArray():
    v2a = ak.contents.unionarray.UnionArray(
        ak.index.Index(np.array([1, 1, 0, 0, 1, 0, 1], np.int8)),
        ak.index.Index(np.array([4, 3, 0, 1, 2, 2, 4, 100], np.int64)),
        [
            ak.from_iter(["1", "2", "3"], highlevel=False),
            ak.contents.numpyarray.NumpyArray(np.array([1.1, 2.2, 3.3, 4.4, 5.5])),
        ],
    )

    layout, report = ak.typetracer.typetracer_with_report(v2a.form_with_key())
    assert len(report.shape_touched) == 0
    tmp = layout.form  # noqa: F841
    assert len(report.shape_touched) == 0


@pytest.mark.parametrize("flatlist_as_rvec", [False, True])
def test_nested_NumpyArray(flatlist_as_rvec):
    v2a = ak.contents.ListOffsetArray(
        ak.index.Index64(np.array([0, 1, 5], dtype=np.int64)),
        ak.contents.numpyarray.NumpyArray(
            np.array([999.0, 0.0, 1.1, 2.2, 3.3]),
            parameters={"some": "stuff", "other": [1, 2, "three"]},
        ),
    )

    layout, report = ak.typetracer.typetracer_with_report(v2a.form_with_key())
    assert len(report.shape_touched) == 0
    tmp = layout.form  # noqa: F841
    assert len(report.shape_touched) == 0


def test_nested_NumpyArray_shape():
    data = np.full((3, 3, 5), 999, dtype=np.int64)
    data[1:3] = np.arange(2 * 3 * 5, dtype=np.int64).reshape(2, 3, 5)

    v2a = ak.contents.ListOffsetArray(
        ak.index.Index64(np.array([0, 1, 3], dtype=np.int64)),
        ak.contents.numpyarray.NumpyArray(data),
    )

    layout, report = ak.typetracer.typetracer_with_report(v2a.form_with_key())
    assert len(report.shape_touched) == 0
    tmp = layout.form  # noqa: F841
    assert len(report.shape_touched) == 1  # this one actually *should* touch


@pytest.mark.parametrize("flatlist_as_rvec", [False, True])
def test_nested_RegularArray_NumpyArray(flatlist_as_rvec):
    v2a = ak.contents.ListOffsetArray(
        ak.index.Index64(np.array([0, 1, 3], dtype=np.int64)),
        ak.contents.regulararray.RegularArray(
            ak.contents.numpyarray.NumpyArray(
                np.array([999, 999, 999, 0.0, 1.1, 2.2, 3.3, 4.4, 5.5])
            ),
            3,
        ),
    )

    layout, report = ak.typetracer.typetracer_with_report(v2a.form_with_key())
    assert len(report.shape_touched) == 0
    tmp = layout.form
    assert len(report.shape_touched) == 0

    v2b = ak.contents.ListOffsetArray(
        ak.index.Index64(np.array([0, 1, 11], dtype=np.int64)),
        ak.contents.regulararray.RegularArray(
            ak.contents.emptyarray.EmptyArray().to_NumpyArray(np.dtype(np.float64)),
            0,
            zeros_length=11,
        ),
    )

    layout, report = ak.typetracer.typetracer_with_report(v2b.form_with_key())
    assert len(report.shape_touched) == 0
    tmp = layout.form  # noqa: F841
    assert len(report.shape_touched) == 0


@pytest.mark.parametrize("flatlist_as_rvec", [False, True])
def test_nested_ListArray_NumpyArray(flatlist_as_rvec):
    v2a = ak.contents.ListOffsetArray(
        ak.index.Index64(np.array([0, 1, 4], dtype=np.int64)),
        ak.contents.listarray.ListArray(
            ak.index.Index(np.array([999, 4, 100, 1], np.int64)),
            ak.index.Index(np.array([999, 7, 100, 3, 200], np.int64)),
            ak.contents.numpyarray.NumpyArray(
                np.array([6.6, 4.4, 5.5, 7.7, 1.1, 2.2, 3.3, 8.8])
            ),
        ),
    )

    layout, report = ak.typetracer.typetracer_with_report(v2a.form_with_key())
    assert len(report.shape_touched) == 0
    tmp = layout.form  # noqa: F841
    assert len(report.shape_touched) == 0


@pytest.mark.parametrize("flatlist_as_rvec", [False, True])
def test_nested_ListOffsetArray_NumpyArray(flatlist_as_rvec):
    v2a = ak.contents.ListOffsetArray(
        ak.index.Index64(np.array([0, 1, 5], dtype=np.int64)),
        ak.contents.listoffsetarray.ListOffsetArray(
            ak.index.Index(np.array([1, 1, 4, 4, 6, 7], np.int64)),
            ak.contents.numpyarray.NumpyArray(
                np.array([6.6, 1.1, 2.2, 3.3, 4.4, 5.5, 7.7])
            ),
        ),
    )

    layout, report = ak.typetracer.typetracer_with_report(v2a.form_with_key())
    assert len(report.shape_touched) == 0
    tmp = layout.form  # noqa: F841
    assert len(report.shape_touched) == 0


def test_nested_RecordArray_NumpyArray():
    v2a = ak.contents.ListOffsetArray(
        ak.index.Index64(np.array([0, 1, 6], dtype=np.int64)),
        ak.contents.recordarray.RecordArray(
            [
                ak.contents.numpyarray.NumpyArray(
                    np.array([999, 0, 1, 2, 3, 4], np.int64)
                ),
                ak.contents.numpyarray.NumpyArray(
                    np.array([999, 0.0, 1.1, 2.2, 3.3, 4.4, 5.5])
                ),
            ],
            ["x", "y"],
            parameters={"__record__": "Something"},
        ),
    )

    layout, report = ak.typetracer.typetracer_with_report(v2a.form_with_key())
    assert len(report.shape_touched) == 0
    tmp = layout.form
    assert len(report.shape_touched) == 0

    v2b = ak.contents.ListOffsetArray(
        ak.index.Index64(np.array([0, 1, 6], dtype=np.int64)),
        ak.contents.recordarray.RecordArray(
            [
                ak.contents.numpyarray.NumpyArray(
                    np.array([999, 0, 1, 2, 3, 4], np.int64)
                ),
                ak.contents.numpyarray.NumpyArray(
                    np.array([999, 0.0, 1.1, 2.2, 3.3, 4.4, 5.5])
                ),
            ],
            None,
        ),
    )

    layout, report = ak.typetracer.typetracer_with_report(v2b.form_with_key())
    assert len(report.shape_touched) == 0
    tmp = layout.form
    assert len(report.shape_touched) == 0

    v2c = ak.contents.ListOffsetArray(
        ak.index.Index64(np.array([0, 1, 11], dtype=np.int64)),
        ak.contents.recordarray.RecordArray([], [], 11),
    )

    layout, report = ak.typetracer.typetracer_with_report(v2c.form_with_key())
    assert len(report.shape_touched) == 0
    tmp = layout.form
    assert len(report.shape_touched) == 0

    v2d = ak.contents.ListOffsetArray(
        ak.index.Index64(np.array([0, 1, 11], dtype=np.int64)),
        ak.contents.recordarray.RecordArray([], None, 11),
    )

    layout, report = ak.typetracer.typetracer_with_report(v2d.form_with_key())
    assert len(report.shape_touched) == 0
    tmp = layout.form  # noqa: F841
    assert len(report.shape_touched) == 0


def test_nested_IndexedArray_NumpyArray():
    v2a = ak.contents.ListOffsetArray(
        ak.index.Index64(np.array([0, 1, 8], dtype=np.int64)),
        ak.contents.indexedarray.IndexedArray(
            ak.index.Index(np.array([999, 2, 2, 0, 1, 4, 5, 4], np.int64)),
            ak.contents.numpyarray.NumpyArray(np.array([0.0, 1.1, 2.2, 3.3, 4.4, 5.5])),
        ),
    )

    layout, report = ak.typetracer.typetracer_with_report(v2a.form_with_key())
    assert len(report.shape_touched) == 0
    tmp = layout.form  # noqa: F841
    assert len(report.shape_touched) == 0


def test_nested_IndexedOptionArray_NumpyArray():
    v2a = ak.contents.ListOffsetArray(
        ak.index.Index64(np.array([0, 1, 8], dtype=np.int64)),
        ak.contents.indexedoptionarray.IndexedOptionArray(
            ak.index.Index(np.array([999, 2, 2, -1, 1, -1, 5, 4], np.int64)),
            ak.contents.numpyarray.NumpyArray(np.array([0.0, 1.1, 2.2, 3.3, 4.4, 5.5])),
        ),
    )

    layout, report = ak.typetracer.typetracer_with_report(v2a.form_with_key())
    assert len(report.shape_touched) == 0
    tmp = layout.form  # noqa: F841
    assert len(report.shape_touched) == 0


def test_nested_ByteMaskedArray_NumpyArray():
    v2a = ak.contents.ListOffsetArray(
        ak.index.Index64(np.array([0, 1, 6], dtype=np.int64)),
        ak.contents.bytemaskedarray.ByteMaskedArray(
            ak.index.Index(np.array([123, 1, 0, 1, 0, 1], np.int8)),
            ak.contents.numpyarray.NumpyArray(
                np.array([999, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6])
            ),
            valid_when=True,
        ),
    )

    layout, report = ak.typetracer.typetracer_with_report(v2a.form_with_key())
    assert len(report.shape_touched) == 0
    tmp = layout.form
    assert len(report.shape_touched) == 0

    v2b = ak.contents.ListOffsetArray(
        ak.index.Index64(np.array([0, 1, 6], dtype=np.int64)),
        ak.contents.bytemaskedarray.ByteMaskedArray(
            ak.index.Index(np.array([123, 0, 1, 0, 1, 0], np.int8)),
            ak.contents.numpyarray.NumpyArray(
                np.array([999, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6])
            ),
            valid_when=False,
        ),
    )

    layout, report = ak.typetracer.typetracer_with_report(v2b.form_with_key())
    assert len(report.shape_touched) == 0
    tmp = layout.form  # noqa: F841
    assert len(report.shape_touched) == 0


def test_nested_BitMaskedArray_NumpyArray():
    v2a = ak.contents.ListOffsetArray(
        ak.index.Index64(np.array([0, 1, 14], dtype=np.int64)),
        ak.contents.bitmaskedarray.BitMaskedArray(
            ak.index.Index(
                np.packbits(
                    np.array(
                        [
                            0,
                            1,
                            1,
                            1,
                            1,
                            0,
                            0,
                            0,
                            0,
                            1,
                            0,
                            1,
                            0,
                            1,
                        ],
                        np.uint8,
                    )
                )
            ),
            ak.contents.numpyarray.NumpyArray(
                np.array(
                    [
                        999,
                        0.0,
                        1.0,
                        2.0,
                        3.0,
                        4.0,
                        5.0,
                        6.0,
                        7.0,
                        1.1,
                        2.2,
                        3.3,
                        4.4,
                        5.5,
                        6.6,
                    ]
                )
            ),
            valid_when=True,
            length=14,
            lsb_order=False,
        ),
    )

    layout, report = ak.typetracer.typetracer_with_report(v2a.form_with_key())
    assert len(report.shape_touched) == 0
    tmp = layout.form
    assert len(report.shape_touched) == 0

    v2b = ak.contents.ListOffsetArray(
        ak.index.Index64(np.array([0, 1, 14], dtype=np.int64)),
        ak.contents.bitmaskedarray.BitMaskedArray(
            ak.index.Index(
                np.packbits(
                    np.array(
                        [
                            0,
                            0,
                            0,
                            0,
                            0,
                            1,
                            1,
                            1,
                            1,
                            0,
                            1,
                            0,
                            1,
                            0,
                        ],
                        np.uint8,
                    )
                )
            ),
            ak.contents.numpyarray.NumpyArray(
                np.array(
                    [
                        999,
                        0.0,
                        1.0,
                        2.0,
                        3.0,
                        4.0,
                        5.0,
                        6.0,
                        7.0,
                        1.1,
                        2.2,
                        3.3,
                        4.4,
                        5.5,
                        6.6,
                    ]
                )
            ),
            valid_when=False,
            length=14,
            lsb_order=False,
        ),
    )

    layout, report = ak.typetracer.typetracer_with_report(v2b.form_with_key())
    assert len(report.shape_touched) == 0
    tmp = layout.form
    assert len(report.shape_touched) == 0

    v2c = ak.contents.ListOffsetArray(
        ak.index.Index64(np.array([0, 1, 14], dtype=np.int64)),
        ak.contents.bitmaskedarray.BitMaskedArray(
            ak.index.Index(
                np.packbits(
                    np.array(
                        [
                            0,
                            0,
                            0,
                            1,
                            1,
                            1,
                            1,
                            0,
                            0,
                            0,
                            1,
                            0,
                            1,
                            0,
                            1,
                            0,
                            0,
                        ],
                        np.uint8,
                    )
                )
            ),
            ak.contents.numpyarray.NumpyArray(
                np.array(
                    [
                        999,
                        0.0,
                        1.0,
                        2.0,
                        3.0,
                        4.0,
                        5.0,
                        6.0,
                        7.0,
                        1.1,
                        2.2,
                        3.3,
                        4.4,
                        5.5,
                        6.6,
                    ]
                )
            ),
            valid_when=True,
            length=14,
            lsb_order=True,
        ),
    )

    layout, report = ak.typetracer.typetracer_with_report(v2c.form_with_key())
    assert len(report.shape_touched) == 0
    tmp = layout.form
    assert len(report.shape_touched) == 0

    v2d = ak.contents.ListOffsetArray(
        ak.index.Index64(np.array([0, 1, 14], dtype=np.int64)),
        ak.contents.bitmaskedarray.BitMaskedArray(
            ak.index.Index(
                np.packbits(
                    np.array(
                        [
                            1,
                            1,
                            1,
                            0,
                            0,
                            0,
                            0,
                            1,
                            1,
                            1,
                            0,
                            1,
                            0,
                            1,
                            0,
                            1,
                            1,
                        ],
                        np.uint8,
                    )
                )
            ),
            ak.contents.numpyarray.NumpyArray(
                np.array(
                    [
                        999,
                        0.0,
                        1.0,
                        2.0,
                        3.0,
                        4.0,
                        5.0,
                        6.0,
                        7.0,
                        1.1,
                        2.2,
                        3.3,
                        4.4,
                        5.5,
                        6.6,
                    ]
                )
            ),
            valid_when=False,
            length=14,
            lsb_order=True,
        ),
    )

    layout, report = ak.typetracer.typetracer_with_report(v2d.form_with_key())
    assert len(report.shape_touched) == 0
    tmp = layout.form  # noqa: F841
    assert len(report.shape_touched) == 0


def test_nested_UnmaskedArray_NumpyArray():
    v2a = ak.contents.ListOffsetArray(
        ak.index.Index64(np.array([0, 1, 5], dtype=np.int64)),
        ak.contents.unmaskedarray.UnmaskedArray(
            ak.contents.numpyarray.NumpyArray(np.array([999, 0.0, 1.1, 2.2, 3.3]))
        ),
    )

    layout, report = ak.typetracer.typetracer_with_report(v2a.form_with_key())
    assert len(report.shape_touched) == 0
    tmp = layout.form  # noqa: F841
    assert len(report.shape_touched) == 0


def test_nested_UnionArray_NumpyArray():
    v2a = ak.contents.ListOffsetArray(
        ak.index.Index64(np.array([0, 1, 8], dtype=np.int64)),
        ak.contents.unionarray.UnionArray(
            ak.index.Index(np.array([123, 1, 1, 0, 0, 1, 0, 1], np.int8)),
            ak.index.Index(np.array([999, 4, 3, 0, 1, 2, 2, 4, 100], np.int64)),
            [
                ak.from_iter(["1", "2", "3"], highlevel=False),
                ak.contents.numpyarray.NumpyArray(np.array([1.1, 2.2, 3.3, 4.4, 5.5])),
            ],
        ),
    )

    layout, report = ak.typetracer.typetracer_with_report(v2a.form_with_key())
    assert len(report.shape_touched) == 0
    tmp = layout.form  # noqa: F841
    assert len(report.shape_touched) == 0


def test_ListArray_strings():
    v2a = ak.operations.from_iter(
        ["one", "two", "three", "four", "five"], highlevel=False
    )

    layout, report = ak.typetracer.typetracer_with_report(v2a.form_with_key())
    assert len(report.shape_touched) == 0
    tmp = layout.form  # noqa: F841
    assert len(report.shape_touched) == 0


def test_RegularArray_strings():
    v2a = ak.operations.to_regular(
        ak.operations.from_iter(["onexx", "twoxx", "three", "fourx", "fivex"]),
        axis=1,
        highlevel=False,
    )

    layout, report = ak.typetracer.typetracer_with_report(v2a.form_with_key())
    assert len(report.shape_touched) == 0
    tmp = layout.form  # noqa: F841
    assert len(report.shape_touched) == 0


def test_NumpyArray_iterator():
    v2a = ak.contents.numpyarray.NumpyArray(
        np.array([0.0, 1.1, 2.2, 3.3]),
        parameters={"some": "stuff", "other": [1, 2, "three"]},
    )

    layout, report = ak.typetracer.typetracer_with_report(v2a.form_with_key())
    assert len(report.shape_touched) == 0
    tmp = layout.form  # noqa: F841
    assert len(report.shape_touched) == 0


def test_NumpyArray_iterator2():
    v2a = ak.contents.numpyarray.NumpyArray(
        np.array([0.0, 1.1, 2.2, 3.3]),
        parameters={"some": "stuff", "other": [1, 2, "three"]},
    )

    layout, report = ak.typetracer.typetracer_with_report(v2a.form_with_key())
    assert len(report.shape_touched) == 0
    tmp = layout.form  # noqa: F841
    assert len(report.shape_touched) == 0


def test_NumpyArray_riterator():
    v2a = ak.contents.numpyarray.NumpyArray(
        np.array([0.0, 1.1, 2.2, 3.3]),
        parameters={"some": "stuff", "other": [1, 2, "three"]},
    )

    layout, report = ak.typetracer.typetracer_with_report(v2a.form_with_key())
    assert len(report.shape_touched) == 0
    tmp = layout.form  # noqa: F841
    assert len(report.shape_touched) == 0



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import numpy as np
import pytest

import awkward as ak

jax = pytest.importorskip("jax")
ak.jax.register_and_check()

# Define all reducers to test
REDUCERS = [
    (ak.argmin, {}),
    (ak.argmax, {}),
    (ak.min, {}),
    (ak.max, {}),
    (ak.sum, {}),
    (ak.prod, {"mask_identity": True}),  # mask_identity for prod to handle empty arrays
    (ak.any, {}),
    (ak.all, {}),
    (ak.count, {}),
    (ak.count_nonzero, {}),
]

# Define test arrays (single jagged)
SINGLE_JAGGED = [
    # Normal array
    [[1, 2, 3], [4, 5], [6, 7, 8, 9]],
    # Array with first empty
    [[], [1, 2], [3, 4, 5]],
    # Array with middle empty
    [[1, 2], [], [3, 4, 5]],
    # Array with last empty
    [[1, 2], [3, 4, 5], []],
    # Array with multiple empty elements
    [[], [1, 2], [], [3, 4], []],
    # Array with negative numbers
    [[-1, -2], [-3], [4, 5, -6]],
    # Array with zeros
    [[0, 0], [1, 0], [2, 3, 4]],
]

# Define test arrays (double jagged)
DOUBLE_JAGGED = [
    # Normal double jagged array
    [[[1, 2], [3]], [[4, 5, 6]], [[7], [8, 9]]],
    # Double jagged with empty at first level
    [[], [[1, 2], [3, 4]], [[5, 6]]],
    # Double jagged with empty at second level
    [[[1, 2], []], [[3, 4], [5]], [[6]]],
    # Double jagged with various empty elements
    [[[]], [[], [1, 2]], [[], [], [3, 4]]],
    # Double jagged with negative numbers
    [[[-1, -2], [-3]], [[4, 5, -6]], [[7], [-8, 9]]],
    # Double jagged with zeros
    [[[0, 0], [1]], [[2, 3]], [[4], [5, 6]]],
]

# Define axes to test
AXES = [1, None]  # axis=1 for first dimension, None for flattened reduction
DOUBLE_JAGGED_AXES = [1, 2, None]  # axis=1 and axis=2 for double jagged

RTOL = 1e-5  # Relative tolerance for floating point comparison
ATOL = 1e-8  # Absolute tolerance for floating point comparison


def compare_results(cpu_list, jax_list):
    """Compare results with tolerance for numeric values."""
    if isinstance(cpu_list, (int, float)) and isinstance(jax_list, (int, float)):
        # Direct numeric comparison with tolerance
        np.testing.assert_allclose(cpu_list, jax_list, rtol=RTOL, atol=ATOL)
    elif isinstance(cpu_list, list) and isinstance(jax_list, list):
        # Lists should have the same length
        assert len(cpu_list) == len(jax_list), (
            f"Lists have different lengths: {len(cpu_list)} vs {len(jax_list)}"
        )

        # Compare each element
        for cpu_item, jax_item in zip(cpu_list, jax_list, strict=True):
            compare_results(cpu_item, jax_item)
    else:
        # For non-numeric types, use exact equality
        assert cpu_list == jax_list


@pytest.mark.parametrize("reducer,kwargs", REDUCERS)
@pytest.mark.parametrize("arr", SINGLE_JAGGED)
@pytest.mark.parametrize("axis", AXES)
def test_single_jagged_arrays(reducer, kwargs, arr, axis):
    """Test reducers on single jagged arrays with different axes."""

    # Create arrays with different backends
    cpu_array = ak.Array(arr, backend="cpu")
    jax_array = ak.Array(arr, backend="jax")

    # Apply reducers to each backend's array
    cpu_result = reducer(cpu_array, axis=axis, **kwargs)
    jax_result = reducer(jax_array, axis=axis, **kwargs)

    # Convert to lists for comparison
    cpu_list = ak.to_list(cpu_result)
    jax_list = ak.to_list(jax_result)

    # Handle case where axis=None might result in different structures
    if axis is None:
        # If one result is a scalar and the other is a list with one element
        if (
            not isinstance(cpu_list, list)
            and isinstance(jax_list, list)
            and len(jax_list) == 1
        ):
            jax_list = jax_list[0]
        elif (
            isinstance(cpu_list, list)
            and not isinstance(jax_list, list)
            and len(cpu_list) == 1
        ):
            cpu_list = cpu_list[0]

    # Compare with tolerance for numeric values
    compare_results(cpu_list, jax_list)


@pytest.mark.parametrize("reducer,kwargs", REDUCERS)
@pytest.mark.parametrize("arr", DOUBLE_JAGGED)
@pytest.mark.parametrize("axis", DOUBLE_JAGGED_AXES)
def test_double_jagged_arrays(reducer, kwargs, arr, axis):
    """Test reducers on double jagged arrays with different axes."""

    # Create arrays with different backends
    cpu_array = ak.Array(arr, backend="cpu")
    jax_array = ak.Array(arr, backend="jax")

    # Apply reducers to each backend's array
    cpu_result = reducer(cpu_array, axis=axis, **kwargs)
    jax_result = reducer(jax_array, axis=axis, **kwargs)

    # Convert to lists for comparison
    cpu_list = ak.to_list(cpu_result)
    jax_list = ak.to_list(jax_result)

    # Handle case where axis=None might result in different structures
    if axis is None:
        # If one result is a scalar and the other is a list with one element
        if (
            not isinstance(cpu_list, list)
            and isinstance(jax_list, list)
            and len(jax_list) == 1
        ):
            jax_list = jax_list[0]
        elif (
            isinstance(cpu_list, list)
            and not isinstance(jax_list, list)
            and len(cpu_list) == 1
        ):
            cpu_list = cpu_list[0]

    # Compare with tolerance for numeric values
    compare_results(cpu_list, jax_list)


# Additional edge cases
@pytest.mark.parametrize("reducer,kwargs", REDUCERS)
def test_all_empty_arrays(reducer, kwargs):
    """Test with arrays that are entirely empty."""

    all_empty_data = [[], [], []]
    cpu_array = ak.Array(all_empty_data, backend="cpu")
    jax_array = ak.Array(all_empty_data, backend="jax")

    cpu_result = reducer(cpu_array, axis=1, **kwargs)
    jax_result = reducer(jax_array, axis=1, **kwargs)

    # Convert to lists for comparison
    cpu_list = ak.to_list(cpu_result)
    jax_list = ak.to_list(jax_result)

    # Handle case where one might be a scalar and the other a list
    if (
        not isinstance(cpu_list, list)
        and isinstance(jax_list, list)
        and len(jax_list) == 1
    ):
        jax_list = jax_list[0]
    elif (
        isinstance(cpu_list, list)
        and not isinstance(jax_list, list)
        and len(cpu_list) == 1
    ):
        cpu_list = cpu_list[0]

    # Compare with tolerance for numeric values
    compare_results(cpu_list, jax_list)


# Test with boolean values
@pytest.mark.parametrize("reducer,kwargs", REDUCERS)
def test_boolean_arrays(reducer, kwargs):
    """Test with boolean arrays."""

    bool_data = [[True, False], [], [True, True, False], [False]]
    cpu_array = ak.Array(bool_data, backend="cpu")
    jax_array = ak.Array(bool_data, backend="jax")

    cpu_result = reducer(cpu_array, axis=1, **kwargs)
    jax_result = reducer(jax_array, axis=1, **kwargs)

    # Convert to lists for comparison
    cpu_list = ak.to_list(cpu_result)
    jax_list = ak.to_list(jax_result)

    # Handle case where one might be a scalar and the other a list
    if (
        not isinstance(cpu_list, list)
        and isinstance(jax_list, list)
        and len(jax_list) == 1
    ):
        jax_list = jax_list[0]
    elif (
        isinstance(cpu_list, list)
        and not isinstance(jax_list, list)
        and len(cpu_list) == 1
    ):
        cpu_list = cpu_list[0]

    # Compare with tolerance for numeric values
    compare_results(cpu_list, jax_list)


# Test with None values
@pytest.mark.parametrize("reducer,kwargs", REDUCERS)
def test_none_arrays(reducer, kwargs):
    """Test with arrays containing None values."""

    none_data = [[None, 1], [2, None], [None, None], [3, 4]]
    cpu_array = ak.Array(none_data, backend="cpu")
    jax_array = ak.Array(none_data, backend="jax")

    cpu_result = reducer(cpu_array, axis=1, **kwargs)
    jax_result = reducer(jax_array, axis=1, **kwargs)

    # Convert to lists for comparison
    cpu_list = ak.to_list(cpu_result)
    jax_list = ak.to_list(jax_result)

    # Handle case where one might be a scalar and the other a list
    if (
        not isinstance(cpu_list, list)
        and isinstance(jax_list, list)
        and len(jax_list) == 1
    ):
        jax_list = jax_list[0]
    elif (
        isinstance(cpu_list, list)
        and not isinstance(jax_list, list)
        and len(cpu_list) == 1
    ):
        cpu_list = cpu_list[0]

    # Compare with tolerance for numeric values
    compare_results(cpu_list, jax_list)


# test with NaN values
@pytest.mark.skip(
    reason="(arg)min/max and any do not work with NaNs in the jax backend"
)
@pytest.mark.parametrize("reducer,kwargs", REDUCERS)
def test_nan_arrays(reducer, kwargs):
    """Test with arrays containing NaN values."""

    nan_data = [[np.nan, 1], [2, np.nan], [np.nan, np.nan], [3, 4]]
    cpu_array = ak.Array(nan_data, backend="cpu")
    jax_array = ak.Array(nan_data, backend="jax")

    cpu_result = reducer(cpu_array, axis=1, **kwargs)
    jax_result = reducer(jax_array, axis=1, **kwargs)

    # Convert to lists for comparison
    cpu_list = ak.to_list(cpu_result)
    jax_list = ak.to_list(jax_result)

    # Handle case where one might be a scalar and the other a list
    if (
        not isinstance(cpu_list, list)
        and isinstance(jax_list, list)
        and len(jax_list) == 1
    ):
        jax_list = jax_list[0]
    elif (
        isinstance(cpu_list, list)
        and not isinstance(jax_list, list)
        and len(cpu_list) == 1
    ):
        cpu_list = cpu_list[0]

    # Compare with tolerance for numeric values
    compare_results(cpu_list, jax_list)



from __future__ import annotations

import cupy as cp
import numpy as np
import pytest

import awkward as ak
from awkward.operations.ak_concatenate import enforce_concatenated_form

to_list = ak.operations.to_list


@pytest.fixture(scope="function", autouse=True)
def cleanup_cuda():
    yield
    try:
        cp.cuda.Device().synchronize()  # wait for all kernels
    except cp.cuda.runtime.CUDARuntimeError as e:
        print("GPU error during sync:", e)
    cp._default_memory_pool.free_all_blocks()


def test_0449_merge_many_arrays_in_one_pass_concatenate():
    one = ak.highlevel.Array([1, 2, 3]).layout
    two = ak.highlevel.Array([4.4, 5.5]).layout
    three = ak.highlevel.Array([6, 7, 8]).layout
    four = ak.highlevel.Array([[9, 9, 9], [10, 10, 10]]).layout

    cuda_one = ak.to_backend(one, "cuda")
    cuda_two = ak.to_backend(two, "cuda")
    cuda_three = ak.to_backend(three, "cuda")
    cuda_four = ak.to_backend(four, "cuda")

    assert to_list(
        ak.operations.concatenate([cuda_one, cuda_two, cuda_three, cuda_four])
    ) == [
        1,
        2,
        3,
        4.4,
        5.5,
        6,
        7,
        8,
        [9, 9, 9],
        [10, 10, 10],
    ]
    assert to_list(
        ak.operations.concatenate([cuda_four, cuda_one, cuda_two, cuda_three])
    ) == [
        [9, 9, 9],
        [10, 10, 10],
        1,
        2,
        3,
        4.4,
        5.5,
        6,
        7,
        8,
    ]
    assert to_list(
        ak.operations.concatenate([cuda_one, cuda_two, cuda_four, cuda_three])
    ) == [
        1,
        2,
        3,
        4.4,
        5.5,
        [9, 9, 9],
        [10, 10, 10],
        6,
        7,
        8,
    ]

    five = ak.highlevel.Array(["nine", "ten"]).layout
    cuda_five = ak.to_backend(five, "cuda")

    assert to_list(
        ak.operations.concatenate([cuda_one, cuda_two, cuda_three, cuda_five])
    ) == [
        1,
        2,
        3,
        4.4,
        5.5,
        6,
        7,
        8,
        "nine",
        "ten",
    ]
    assert to_list(
        ak.operations.concatenate([cuda_five, cuda_one, cuda_two, cuda_three])
    ) == [
        "nine",
        "ten",
        1,
        2,
        3,
        4.4,
        5.5,
        6,
        7,
        8,
    ]
    assert to_list(
        ak.operations.concatenate([cuda_one, cuda_two, cuda_five, cuda_three])
    ) == [
        1,
        2,
        3,
        4.4,
        5.5,
        "nine",
        "ten",
        6,
        7,
        8,
    ]


def test_0449_merge_many_arrays_in_one_pass_lists():
    one = ak.highlevel.Array([[1, 2, 3], [], [4, 5]]).layout
    two = ak.highlevel.Array([[1.1, 2.2], [3.3, 4.4]]).layout
    three = ak.contents.EmptyArray()
    four = ak.operations.from_numpy(
        np.array([[10], [20]]), regulararray=True, highlevel=False
    )

    cuda_one = ak.to_backend(one, "cuda", highlevel=False)
    cuda_two = ak.to_backend(two, "cuda", highlevel=False)
    cuda_three = ak.to_backend(three, "cuda", highlevel=False)
    cuda_four = ak.to_backend(four, "cuda", highlevel=False)

    assert to_list(cuda_one._mergemany([cuda_two, cuda_three, cuda_four])) == [
        [1.0, 2.0, 3.0],
        [],
        [4.0, 5.0],
        [1.1, 2.2],
        [3.3, 4.4],
        [10.0],
        [20.0],
    ]
    assert to_list(cuda_four._mergemany([cuda_three, cuda_two, cuda_one])) == [
        [10.0],
        [20.0],
        [1.1, 2.2],
        [3.3, 4.4],
        [1.0, 2.0, 3.0],
        [],
        [4.0, 5.0],
    ]
    assert (
        cuda_one.to_typetracer()
        ._mergemany(
            [
                cuda_two.to_typetracer(),
                cuda_three.to_typetracer(),
                cuda_four.to_typetracer(),
            ]
        )
        .form
        == cuda_one._mergemany([cuda_two, cuda_three, cuda_four]).form
    )
    assert (
        cuda_four.to_typetracer()
        ._mergemany(
            [
                cuda_three.to_typetracer(),
                cuda_four.to_typetracer(),
                cuda_one.to_typetracer(),
            ]
        )
        .form
        == cuda_four._mergemany([cuda_three, cuda_four, cuda_one]).form
    )

    one = ak.highlevel.Array([[1, 2, 3], [], [4, 5]]).layout
    two = ak.highlevel.Array([[1.1, 2.2], [3.3, 4.4]]).layout
    one = ak.contents.ListArray(one.starts, one.stops, one.content)
    two = ak.contents.ListArray(two.starts, two.stops, two.content)
    cuda_one = ak.to_backend(one, "cuda", highlevel=False)
    cuda_two = ak.to_backend(two, "cuda", highlevel=False)

    assert to_list(cuda_one._mergemany([cuda_two, cuda_three, cuda_four])) == [
        [1.0, 2.0, 3.0],
        [],
        [4.0, 5.0],
        [1.1, 2.2],
        [3.3, 4.4],
        [10.0],
        [20.0],
    ]
    assert to_list(cuda_four._mergemany([cuda_three, cuda_two, cuda_one])) == [
        [10.0],
        [20.0],
        [1.1, 2.2],
        [3.3, 4.4],
        [1.0, 2.0, 3.0],
        [],
        [4.0, 5.0],
    ]

    assert (
        cuda_one.to_typetracer()
        ._mergemany(
            [
                cuda_two.to_typetracer(),
                cuda_three.to_typetracer(),
                cuda_four.to_typetracer(),
            ]
        )
        .form
        == cuda_one._mergemany([cuda_two, cuda_three, cuda_four]).form
    )

    assert (
        cuda_four.to_typetracer()
        ._mergemany(
            [
                cuda_three.to_typetracer(),
                cuda_four.to_typetracer(),
                cuda_one.to_typetracer(),
            ]
        )
        .form
        == cuda_four._mergemany([cuda_three, cuda_four, cuda_one]).form
    )


def test_0449_merge_many_arrays_in_one_pass_records():
    one = ak.highlevel.Array(
        [{"x": 1, "y": [1]}, {"x": 2, "y": [1, 2]}, {"x": 3, "y": [1, 2, 3]}]
    ).layout
    two = ak.highlevel.Array([{"y": [], "x": 4}, {"y": [3, 2, 1], "x": 5}]).layout
    four = ak.highlevel.Array([{"x": 6, "y": [1]}, {"x": 7, "y": [1, 2]}]).layout

    cuda_one = ak.to_backend(one, "cuda", highlevel=False)
    cuda_two = ak.to_backend(two, "cuda", highlevel=False)
    cuda_three = cuda_two[0:0]
    cuda_four = ak.to_backend(four, "cuda", highlevel=False)

    assert to_list(cuda_one._mergemany([cuda_two, cuda_three, cuda_four])) == [
        {"x": 1, "y": [1]},
        {"x": 2, "y": [1, 2]},
        {"x": 3, "y": [1, 2, 3]},
        {"y": [], "x": 4},
        {"y": [3, 2, 1], "x": 5},
        {"x": 6, "y": [1]},
        {"x": 7, "y": [1, 2]},
    ]
    assert (
        cuda_one.to_typetracer()
        ._mergemany(
            [
                cuda_two.to_typetracer(),
                cuda_three.to_typetracer(),
                cuda_four.to_typetracer(),
            ]
        )
        .form
        == cuda_one._mergemany([cuda_two, cuda_three, cuda_four]).form
    )

    three = ak.contents.EmptyArray()
    cuda_three = ak.to_backend(three, "cuda", highlevel=False)

    assert to_list(cuda_one._mergemany([cuda_two, cuda_three, cuda_four])) == [
        {"x": 1, "y": [1]},
        {"x": 2, "y": [1, 2]},
        {"x": 3, "y": [1, 2, 3]},
        {"y": [], "x": 4},
        {"y": [3, 2, 1], "x": 5},
        {"x": 6, "y": [1]},
        {"x": 7, "y": [1, 2]},
    ]
    assert (
        cuda_one.to_typetracer()
        ._mergemany(
            [
                cuda_two.to_typetracer(),
                cuda_three.to_typetracer(),
                cuda_four.to_typetracer(),
            ]
        )
        .form
        == cuda_one._mergemany([cuda_two, cuda_three, cuda_four]).form
    )


def test_0449_merge_many_arrays_in_one_pass_tuples():
    one = ak.highlevel.Array([(1, [1]), (2, [1, 2]), (3, [1, 2, 3])]).layout
    two = ak.highlevel.Array([(4, []), (5, [3, 2, 1])]).layout
    four = ak.highlevel.Array([(6, [1]), (7, [1, 2])]).layout

    cuda_one = ak.to_backend(one, "cuda", highlevel=False)
    cuda_two = ak.to_backend(two, "cuda", highlevel=False)
    cuda_three = cuda_two[0:0]
    cuda_four = ak.to_backend(four, "cuda", highlevel=False)

    assert to_list(cuda_one._mergemany([cuda_two, cuda_three, cuda_four])) == [
        (1, [1]),
        (2, [1, 2]),
        (3, [1, 2, 3]),
        (4, []),
        (5, [3, 2, 1]),
        (6, [1]),
        (7, [1, 2]),
    ]
    assert (
        cuda_one.to_typetracer()
        ._mergemany(
            [
                cuda_two.to_typetracer(),
                cuda_three.to_typetracer(),
                cuda_four.to_typetracer(),
            ]
        )
        .form
        == cuda_one._mergemany([cuda_two, cuda_three, cuda_four]).form
    )

    three = ak.contents.EmptyArray()
    cuda_three = ak.to_backend(three, "cuda", highlevel=False)

    assert to_list(cuda_one._mergemany([cuda_two, cuda_three, cuda_four])) == [
        (1, [1]),
        (2, [1, 2]),
        (3, [1, 2, 3]),
        (4, []),
        (5, [3, 2, 1]),
        (6, [1]),
        (7, [1, 2]),
    ]
    assert (
        cuda_one.to_typetracer()
        ._mergemany(
            [
                cuda_two.to_typetracer(),
                cuda_three.to_typetracer(),
                cuda_four.to_typetracer(),
            ]
        )
        .form
        == cuda_one._mergemany([cuda_two, cuda_three, cuda_four]).form
    )


def test_0449_merge_many_arrays_in_one_pass_indexed():
    one = ak.highlevel.Array([1, 2, 3, None, 4, None, None, 5]).layout
    two = ak.highlevel.Array([6, 7, 8]).layout
    three = ak.contents.EmptyArray()
    four = ak.highlevel.Array([9, None, None]).layout

    cuda_one = ak.to_backend(one, "cuda", highlevel=False)
    cuda_two = ak.to_backend(two, "cuda", highlevel=False)
    cuda_three = ak.to_backend(three, "cuda", highlevel=False)
    cuda_four = ak.to_backend(four, "cuda", highlevel=False)

    assert to_list(cuda_one._mergemany([cuda_two, cuda_three, cuda_four])) == [
        1,
        2,
        3,
        None,
        4,
        None,
        None,
        5,
        6,
        7,
        8,
        9,
        None,
        None,
    ]
    assert (
        cuda_one.to_typetracer()
        ._mergemany(
            [
                cuda_two.to_typetracer(),
                cuda_three.to_typetracer(),
                cuda_four.to_typetracer(),
            ]
        )
        .form
        == cuda_one._mergemany([cuda_two, cuda_three, cuda_four]).form
    )


def test_0449_merge_many_arrays_in_one_pass_reverse_indexed():
    one = ak.highlevel.Array([1, 2, 3]).layout
    two = ak.highlevel.Array([4, 5]).layout
    three = ak.highlevel.Array([None, 6, None]).layout

    cuda_one = ak.to_backend(one, "cuda", highlevel=False)
    cuda_two = ak.to_backend(two, "cuda", highlevel=False)
    cuda_three = ak.to_backend(three, "cuda", highlevel=False)

    assert to_list(cuda_one._mergemany([cuda_two, cuda_three])) == [
        1,
        2,
        3,
        4,
        5,
        None,
        6,
        None,
    ]
    assert (
        cuda_one.to_typetracer()
        ._mergemany([cuda_two.to_typetracer(), cuda_three.to_typetracer()])
        .form
        == cuda_one._mergemany([cuda_two, cuda_three]).form
    )

    four = ak.highlevel.Array([7, 8, None, None, 9]).layout
    cuda_four = ak.to_backend(four, "cuda", highlevel=False)

    assert to_list(cuda_one._mergemany([cuda_two, cuda_three, cuda_four])) == [
        1,
        2,
        3,
        4,
        5,
        None,
        6,
        None,
        7,
        8,
        None,
        None,
        9,
    ]
    assert (
        cuda_one.to_typetracer()
        ._mergemany(
            [
                cuda_two.to_typetracer(),
                cuda_three.to_typetracer(),
                cuda_four.to_typetracer(),
            ]
        )
        .form
        == cuda_one._mergemany([cuda_two, cuda_three, cuda_four]).form
    )


def test_0449_merge_many_arrays_in_one_pass_bytemasked():
    one = ak.contents.ByteMaskedArray(
        ak.index.Index8([True, True, False, True, False, True]),
        ak.highlevel.Array([1, 2, 3, 4, 5, 6]).layout,
        valid_when=True,
    )
    two = ak.contents.ByteMaskedArray(
        ak.index.Index8([True, False, False, True, True]),
        ak.highlevel.Array([7, 99, 999, 8, 9]).layout,
        valid_when=True,
    )
    three = ak.highlevel.Array([100, 200, 300]).layout
    four = ak.highlevel.Array([None, None, 123, None]).layout

    cuda_one = ak.to_backend(one, "cuda", highlevel=False)
    cuda_two = ak.to_backend(two, "cuda", highlevel=False)
    cuda_three = ak.to_backend(three, "cuda", highlevel=False)
    cuda_four = ak.to_backend(four, "cuda", highlevel=False)

    assert to_list(cuda_one._mergemany([cuda_two, cuda_three, cuda_four])) == [
        1,
        2,
        None,
        4,
        None,
        6,
        7,
        None,
        None,
        8,
        9,
        100,
        200,
        300,
        None,
        None,
        123,
        None,
    ]
    assert to_list(cuda_four._mergemany([cuda_three, cuda_two, cuda_one])) == [
        None,
        None,
        123,
        None,
        100,
        200,
        300,
        7,
        None,
        None,
        8,
        9,
        1,
        2,
        None,
        4,
        None,
        6,
    ]
    assert to_list(cuda_three._mergemany([cuda_four, cuda_one])) == [
        100,
        200,
        300,
        None,
        None,
        123,
        None,
        1,
        2,
        None,
        4,
        None,
        6,
    ]
    assert to_list(cuda_three._mergemany([cuda_four, cuda_one, cuda_two])) == [
        100,
        200,
        300,
        None,
        None,
        123,
        None,
        1,
        2,
        None,
        4,
        None,
        6,
        7,
        None,
        None,
        8,
        9,
    ]
    assert to_list(cuda_three._mergemany([cuda_two, cuda_one])) == [
        100,
        200,
        300,
        7,
        None,
        None,
        8,
        9,
        1,
        2,
        None,
        4,
        None,
        6,
    ]
    assert to_list(cuda_three._mergemany([cuda_two, cuda_one, cuda_four])) == [
        100,
        200,
        300,
        7,
        None,
        None,
        8,
        9,
        1,
        2,
        None,
        4,
        None,
        6,
        None,
        None,
        123,
        None,
    ]

    assert (
        cuda_one.to_typetracer()
        ._mergemany(
            [
                cuda_two.to_typetracer(),
                cuda_three.to_typetracer(),
                cuda_four.to_typetracer(),
            ]
        )
        .form
        == cuda_one._mergemany([cuda_two, cuda_three, cuda_four]).form
    )
    assert (
        cuda_four.to_typetracer()
        ._mergemany(
            [
                cuda_three.to_typetracer(),
                cuda_two.to_typetracer(),
                cuda_one.to_typetracer(),
            ]
        )
        .form
        == cuda_four._mergemany([cuda_three, cuda_two, cuda_one]).form
    )
    assert (
        cuda_three.to_typetracer()
        ._mergemany([cuda_four.to_typetracer(), cuda_one.to_typetracer()])
        .form
        == cuda_three._mergemany([cuda_four, cuda_one]).form
    )
    assert (
        cuda_three.to_typetracer()
        ._mergemany(
            [
                cuda_four.to_typetracer(),
                cuda_one.to_typetracer(),
                cuda_two.to_typetracer(),
            ]
        )
        .form
        == cuda_three._mergemany([cuda_four, cuda_one, cuda_two]).form
    )
    assert (
        cuda_three.to_typetracer()
        ._mergemany([cuda_two.to_typetracer(), cuda_one.to_typetracer()])
        .form
        == cuda_three._mergemany([cuda_two, cuda_one]).form
    )
    assert (
        cuda_three.to_typetracer()
        ._mergemany(
            [
                cuda_two.to_typetracer(),
                cuda_one.to_typetracer(),
                cuda_four.to_typetracer(),
            ]
        )
        .form
        == cuda_three._mergemany([cuda_two, cuda_one, cuda_four]).form
    )


def test_0449_merge_many_arrays_in_one_pass_union():
    one = ak.highlevel.Array([1, 2, [], [3, 4]]).layout
    two = ak.highlevel.Array([100, 200, 300]).layout
    three = ak.highlevel.Array([{"x": 1}, {"x": 2}, 5, 6, 7]).layout

    cuda_one = ak.to_backend(one, "cuda", highlevel=False)
    cuda_two = ak.to_backend(two, "cuda", highlevel=False)
    cuda_three = ak.to_backend(three, "cuda", highlevel=False)

    assert to_list(cuda_one._mergemany([cuda_two, cuda_three])) == [
        1,
        2,
        [],
        [3, 4],
        100,
        200,
        300,
        {"x": 1},
        {"x": 2},
        5,
        6,
        7,
    ]
    assert to_list(cuda_one._mergemany([cuda_three, cuda_two])) == [
        1,
        2,
        [],
        [3, 4],
        {"x": 1},
        {"x": 2},
        5,
        6,
        7,
        100,
        200,
        300,
    ]
    assert to_list(cuda_two._mergemany([cuda_one, cuda_three])) == [
        100,
        200,
        300,
        1,
        2,
        [],
        [3, 4],
        {"x": 1},
        {"x": 2},
        5,
        6,
        7,
    ]
    assert to_list(cuda_two._mergemany([cuda_three, cuda_one])) == [
        100,
        200,
        300,
        {"x": 1},
        {"x": 2},
        5,
        6,
        7,
        1,
        2,
        [],
        [3, 4],
    ]
    assert to_list(cuda_three._mergemany([cuda_one, cuda_two])) == [
        {"x": 1},
        {"x": 2},
        5,
        6,
        7,
        1,
        2,
        [],
        [3, 4],
        100,
        200,
        300,
    ]
    assert to_list(cuda_three._mergemany([cuda_two, cuda_one])) == [
        {"x": 1},
        {"x": 2},
        5,
        6,
        7,
        100,
        200,
        300,
        1,
        2,
        [],
        [3, 4],
    ]

    assert (
        cuda_one.to_typetracer()
        ._mergemany([cuda_two.to_typetracer(), cuda_three.to_typetracer()])
        .form
        == cuda_one._mergemany([cuda_two, cuda_three]).form
    )
    assert (
        cuda_one.to_typetracer()
        ._mergemany([cuda_three.to_typetracer(), cuda_two.to_typetracer()])
        .form
        == cuda_one._mergemany([cuda_three, cuda_two]).form
    )
    assert (
        cuda_two.to_typetracer()
        ._mergemany([cuda_one.to_typetracer(), cuda_three.to_typetracer()])
        .form
        == cuda_two._mergemany([cuda_one, cuda_three]).form
    )
    assert (
        cuda_two.to_typetracer()
        ._mergemany([cuda_three.to_typetracer(), cuda_one.to_typetracer()])
        .form
        == cuda_two._mergemany([cuda_three, cuda_one]).form
    )
    assert (
        cuda_three.to_typetracer()
        ._mergemany([cuda_one.to_typetracer(), cuda_two.to_typetracer()])
        .form
        == cuda_three._mergemany([cuda_one, cuda_two]).form
    )
    assert (
        cuda_three.to_typetracer()
        ._mergemany([cuda_two.to_typetracer(), cuda_one.to_typetracer()])
        .form
        == cuda_three._mergemany([cuda_two, cuda_one]).form
    )


def test_0449_merge_many_arrays_in_one_pass_union_option():
    one = ak.highlevel.Array([1, 2, [], [3, 4]]).layout
    two = ak.highlevel.Array([100, None, 300]).layout
    three = ak.highlevel.Array([{"x": 1}, {"x": 2}, 5, 6, 7]).layout

    cuda_one = ak.to_backend(one, "cuda", highlevel=False)
    cuda_two = ak.to_backend(two, "cuda", highlevel=False)
    cuda_three = ak.to_backend(three, "cuda", highlevel=False)

    assert to_list(cuda_one._mergemany([cuda_two, cuda_three])) == [
        1,
        2,
        [],
        [3, 4],
        100,
        None,
        300,
        {"x": 1},
        {"x": 2},
        5,
        6,
        7,
    ]
    assert to_list(cuda_one._mergemany([cuda_three, cuda_two])) == [
        1,
        2,
        [],
        [3, 4],
        {"x": 1},
        {"x": 2},
        5,
        6,
        7,
        100,
        None,
        300,
    ]
    assert to_list(cuda_two._mergemany([cuda_one, cuda_three])) == [
        100,
        None,
        300,
        1,
        2,
        [],
        [3, 4],
        {"x": 1},
        {"x": 2},
        5,
        6,
        7,
    ]
    assert to_list(cuda_two._mergemany([cuda_three, cuda_one])) == [
        100,
        None,
        300,
        {"x": 1},
        {"x": 2},
        5,
        6,
        7,
        1,
        2,
        [],
        [3, 4],
    ]
    assert to_list(cuda_three._mergemany([cuda_one, cuda_two])) == [
        {"x": 1},
        {"x": 2},
        5,
        6,
        7,
        1,
        2,
        [],
        [3, 4],
        100,
        None,
        300,
    ]
    assert to_list(cuda_three._mergemany([cuda_two, cuda_one])) == [
        {"x": 1},
        {"x": 2},
        5,
        6,
        7,
        100,
        None,
        300,
        1,
        2,
        [],
        [3, 4],
    ]

    assert (
        cuda_one.to_typetracer()
        ._mergemany([cuda_two.to_typetracer(), cuda_three.to_typetracer()])
        .form
        == cuda_one._mergemany([cuda_two, cuda_three]).form
    )
    assert (
        cuda_one.to_typetracer()
        ._mergemany([cuda_three.to_typetracer(), cuda_two.to_typetracer()])
        .form
        == cuda_one._mergemany([cuda_three, cuda_two]).form
    )
    assert (
        cuda_two.to_typetracer()
        ._mergemany([cuda_one.to_typetracer(), cuda_three.to_typetracer()])
        .form
        == cuda_two._mergemany([cuda_one, cuda_three]).form
    )
    assert (
        cuda_two.to_typetracer()
        ._mergemany([cuda_three.to_typetracer(), cuda_one.to_typetracer()])
        .form
        == cuda_two._mergemany([cuda_three, cuda_one]).form
    )
    assert (
        cuda_three.to_typetracer()
        ._mergemany([cuda_one.to_typetracer(), cuda_two.to_typetracer()])
        .form
        == cuda_three._mergemany([cuda_one, cuda_two]).form
    )
    assert (
        cuda_three.to_typetracer()
        ._mergemany([cuda_two.to_typetracer(), cuda_one.to_typetracer()])
        .form
        == cuda_three._mergemany([cuda_two, cuda_one]).form
    )

    one = ak.highlevel.Array([1, 2, [], [3, 4]]).layout
    two = ak.highlevel.Array([100, None, 300]).layout
    three = ak.highlevel.Array([{"x": 1}, {"x": 2}, 5, None, 7]).layout

    cuda_one = ak.to_backend(one, "cuda", highlevel=False)
    cuda_two = ak.to_backend(two, "cuda", highlevel=False)
    cuda_three = ak.to_backend(three, "cuda", highlevel=False)

    assert to_list(cuda_one._mergemany([cuda_two, cuda_three])) == [
        1,
        2,
        [],
        [3, 4],
        100,
        None,
        300,
        {"x": 1},
        {"x": 2},
        5,
        None,
        7,
    ]
    assert to_list(cuda_one._mergemany([cuda_three, cuda_two])) == [
        1,
        2,
        [],
        [3, 4],
        {"x": 1},
        {"x": 2},
        5,
        None,
        7,
        100,
        None,
        300,
    ]
    assert to_list(cuda_two._mergemany([cuda_one, cuda_three])) == [
        100,
        None,
        300,
        1,
        2,
        [],
        [3, 4],
        {"x": 1},
        {"x": 2},
        5,
        None,
        7,
    ]
    assert to_list(cuda_two._mergemany([cuda_three, cuda_one])) == [
        100,
        None,
        300,
        {"x": 1},
        {"x": 2},
        5,
        None,
        7,
        1,
        2,
        [],
        [3, 4],
    ]
    assert to_list(cuda_three._mergemany([cuda_one, cuda_two])) == [
        {"x": 1},
        {"x": 2},
        5,
        None,
        7,
        1,
        2,
        [],
        [3, 4],
        100,
        None,
        300,
    ]
    assert to_list(cuda_three._mergemany([cuda_two, cuda_one])) == [
        {"x": 1},
        {"x": 2},
        5,
        None,
        7,
        100,
        None,
        300,
        1,
        2,
        [],
        [3, 4],
    ]

    assert (
        cuda_one.to_typetracer()
        ._mergemany([cuda_two.to_typetracer(), cuda_three.to_typetracer()])
        .form
        == cuda_one._mergemany([cuda_two, cuda_three]).form
    )
    assert (
        cuda_one.to_typetracer()
        ._mergemany([cuda_three.to_typetracer(), cuda_two.to_typetracer()])
        .form
        == cuda_one._mergemany([cuda_three, cuda_two]).form
    )
    assert (
        cuda_two.to_typetracer()
        ._mergemany([cuda_one.to_typetracer(), cuda_three.to_typetracer()])
        .form
        == cuda_two._mergemany([cuda_one, cuda_three]).form
    )
    assert (
        cuda_two.to_typetracer()
        ._mergemany([cuda_three.to_typetracer(), cuda_one.to_typetracer()])
        .form
        == cuda_two._mergemany([cuda_three, cuda_one]).form
    )
    assert (
        cuda_three.to_typetracer()
        ._mergemany([cuda_one.to_typetracer(), cuda_two.to_typetracer()])
        .form
        == cuda_three._mergemany([cuda_one, cuda_two]).form
    )
    assert (
        cuda_three.to_typetracer()
        ._mergemany([cuda_two.to_typetracer(), cuda_one.to_typetracer()])
        .form
        == cuda_three._mergemany([cuda_two, cuda_one]).form
    )

    one = ak.highlevel.Array([1, 2, [], [3, 4]]).layout
    two = ak.highlevel.Array([100, 200, 300]).layout
    three = ak.highlevel.Array([{"x": 1}, {"x": 2}, 5, None, 7]).layout

    cuda_one = ak.to_backend(one, "cuda", highlevel=False)
    cuda_two = ak.to_backend(two, "cuda", highlevel=False)
    cuda_three = ak.to_backend(three, "cuda", highlevel=False)

    assert to_list(cuda_one._mergemany([cuda_two, cuda_three])) == [
        1,
        2,
        [],
        [3, 4],
        100,
        200,
        300,
        {"x": 1},
        {"x": 2},
        5,
        None,
        7,
    ]
    assert to_list(cuda_one._mergemany([cuda_three, cuda_two])) == [
        1,
        2,
        [],
        [3, 4],
        {"x": 1},
        {"x": 2},
        5,
        None,
        7,
        100,
        200,
        300,
    ]
    assert to_list(cuda_two._mergemany([cuda_one, cuda_three])) == [
        100,
        200,
        300,
        1,
        2,
        [],
        [3, 4],
        {"x": 1},
        {"x": 2},
        5,
        None,
        7,
    ]
    assert to_list(cuda_two._mergemany([cuda_three, cuda_one])) == [
        100,
        200,
        300,
        {"x": 1},
        {"x": 2},
        5,
        None,
        7,
        1,
        2,
        [],
        [3, 4],
    ]
    assert to_list(cuda_three._mergemany([cuda_one, cuda_two])) == [
        {"x": 1},
        {"x": 2},
        5,
        None,
        7,
        1,
        2,
        [],
        [3, 4],
        100,
        200,
        300,
    ]
    assert to_list(cuda_three._mergemany([cuda_two, cuda_one])) == [
        {"x": 1},
        {"x": 2},
        5,
        None,
        7,
        100,
        200,
        300,
        1,
        2,
        [],
        [3, 4],
    ]

    assert (
        cuda_one.to_typetracer()
        ._mergemany([cuda_two.to_typetracer(), cuda_three.to_typetracer()])
        .form
        == cuda_one._mergemany([cuda_two, cuda_three]).form
    )
    assert (
        cuda_one.to_typetracer()
        ._mergemany([cuda_three.to_typetracer(), cuda_two.to_typetracer()])
        .form
        == cuda_one._mergemany([cuda_three, cuda_two]).form
    )
    assert (
        cuda_two.to_typetracer()
        ._mergemany([cuda_one.to_typetracer(), cuda_three.to_typetracer()])
        .form
        == cuda_two._mergemany([cuda_one, cuda_three]).form
    )
    assert (
        cuda_two.to_typetracer()
        ._mergemany([cuda_three.to_typetracer(), cuda_one.to_typetracer()])
        .form
        == cuda_two._mergemany([cuda_three, cuda_one]).form
    )
    assert (
        cuda_three.to_typetracer()
        ._mergemany([cuda_one.to_typetracer(), cuda_two.to_typetracer()])
        .form
        == cuda_three._mergemany([cuda_one, cuda_two]).form
    )
    assert (
        cuda_three.to_typetracer()
        ._mergemany([cuda_two.to_typetracer(), cuda_one.to_typetracer()])
        .form
        == cuda_three._mergemany([cuda_two, cuda_one]).form
    )


def test_0449_merge_many_arrays_in_one_pass_strings():
    one = ak.highlevel.Array(["uno", "dos", "tres"]).layout
    two = ak.highlevel.Array(["un", "deux", "trois", "quatre"]).layout
    three = ak.highlevel.Array(["onay", "ootay", "eethray"]).layout

    cuda_one = ak.to_backend(one, "cuda", highlevel=False)
    cuda_two = ak.to_backend(two, "cuda", highlevel=False)
    cuda_three = ak.to_backend(three, "cuda", highlevel=False)

    assert to_list(cuda_one._mergemany([cuda_two, cuda_three])) == [
        "uno",
        "dos",
        "tres",
        "un",
        "deux",
        "trois",
        "quatre",
        "onay",
        "ootay",
        "eethray",
    ]

    assert (
        cuda_one.to_typetracer()
        ._mergemany([cuda_two.to_typetracer(), cuda_three.to_typetracer()])
        .form
        == cuda_one._mergemany([cuda_two, cuda_three]).form
    )


layouts = [
    # ListArray
    ak.contents.ListArray(
        ak.index.Index64([0, 3]),
        ak.index.Index64([3, 6]),
        ak.contents.NumpyArray(np.arange(6, dtype=np.int64)),
    ),
    # ListArray
    ak.contents.ListOffsetArray(
        ak.index.Index64([0, 3, 6]),
        ak.contents.NumpyArray(np.arange(6, dtype=np.int64)),
    ),
    # RegularArray
    ak.contents.RegularArray(ak.contents.NumpyArray(np.arange(6, dtype=np.int64)), 3),
    ak.contents.RegularArray(ak.contents.NumpyArray(np.arange(6, dtype=np.int64)), 2),
    # ByteMaskedArray
    ak.contents.ByteMaskedArray(
        ak.index.Index8([True, False, False, True]),
        ak.contents.NumpyArray(np.arange(6, dtype=np.int32)),
        valid_when=True,
    ),
    # ByteMaskedArray
    ak.contents.BitMaskedArray(
        ak.index.IndexU8([1 << 0 | 1 << 1 | 0 << 2 | 0 << 3 | 1 << 4 | 0 << 5]),
        ak.contents.NumpyArray(np.arange(6, dtype=np.int32)),
        valid_when=True,
        lsb_order=True,
        length=6,
    ),
    # UnmaskedArray
    ak.contents.UnmaskedArray(ak.contents.NumpyArray(np.arange(6, dtype=np.int32))),
    # IndexedOptionArray
    ak.contents.IndexedOptionArray(
        ak.index.Index64([3, 1, -1, -1, 2, 0, -1]),
        ak.contents.NumpyArray(np.arange(6, dtype=np.int32)),
    ),
    # NumpyArray
    ak.contents.NumpyArray(np.arange(6, dtype=np.int16)),
    ak.contents.NumpyArray(np.arange(6 * 4, dtype=np.float32).reshape(6, 4)),
    # IndexedArray
    ak.contents.IndexedArray(
        ak.index.Index64([3, 1, 1, 0, 2, 0, 0]),
        ak.contents.NumpyArray(np.arange(6, dtype=np.int32)),
    ),
    # RecordArray
    ak.contents.RecordArray(
        [ak.contents.NumpyArray(np.arange(6, dtype=np.int16))], ["x"]
    ),
    ak.contents.RecordArray(
        [ak.contents.NumpyArray(np.arange(6, dtype=np.float64))], ["y"]
    ),
    ak.contents.RecordArray(
        [ak.contents.NumpyArray(np.arange(6, dtype=np.float32))], None
    ),
    # UnionArray
    ak.contents.UnionArray(
        ak.index.Index8([0, 0, 1]),
        ak.index.Index64([0, 1, 0]),
        [
            ak.contents.NumpyArray(np.arange(6, dtype=np.int16)),
            ak.contents.RecordArray(
                [ak.contents.NumpyArray(np.arange(6, dtype=np.float32))], None
            ),
        ],
    ),
]
cuda_layouts = ak.to_backend(layouts, "cuda", highlevel=False)


@pytest.mark.parametrize("left", cuda_layouts)
@pytest.mark.parametrize("right", cuda_layouts)
def test_2860_enforce_concatenated_form_symmetric(left, right):
    result = ak.concatenate([left, right], axis=0, highlevel=False)
    part_0_result = enforce_concatenated_form(left, result.form)
    assert part_0_result.form == result.form

    part_1_result = enforce_concatenated_form(right, result.form)
    assert part_1_result.form == result.form

    assert part_0_result.to_list() == result[: part_0_result.length].to_list()
    assert part_1_result.to_list() == result[part_0_result.length :].to_list()
    assert (
        ak.backend(result)
        == ak.backend(part_0_result)
        == ak.backend(part_1_result)
        == "cuda"
    )



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import awkward as ak
from awkward._dispatch import high_level_function
from awkward._nplikes.numpy_like import NumpyMetadata

__all__ = ("to_raggedtensor",)

np = NumpyMetadata.instance()


@high_level_function()
def to_raggedtensor(array):
    """
    Args:
        array: Array-like data (anything #ak.to_layout recognizes).

    Converts `array` (only ListOffsetArray, ListArray, RegularArray and NumpyArray data types supported)
    into a ragged tensor, if possible.

    If `array` contains any other data types (RecordArray for example) the function raises an error.
    """

    # Dispatch
    yield (array,)

    # Implementation
    return _impl(array)


def _impl(array):
    try:
        import tensorflow as tf
    except ImportError as err:
        raise ImportError(
            """to use ak.to_raggedtensor, you must install the 'tensorflow' package with:

        pip install tensorflow
or
        conda install tensorflow"""
        ) from err

    # unwrap the awkward array if it was made with ak.Array function
    # also transforms a python list to awkward array
    array = ak.to_layout(
        ak.operations.materialize(array)
        if isinstance(array, (ak.highlevel.Array, ak.contents.Content))
        else array,
        allow_record=False,
    )

    # keep the same device
    ak_device = ak.backend(array)
    if ak_device not in ["cuda", "cpu"]:
        raise ValueError("""Only 'cpu' and 'cuda' backend conversions are allowed""")

    if ak_device == "cpu":
        device = "CPU:0"
    else:
        id = _find_innermost_content(array).data.device.id
        device = "GPU:" + str(id)

    with tf.device(device):
        if isinstance(array, ak.contents.numpyarray.NumpyArray):
            values = array.data
            # handle cupy separately
            values = _convert_to_tensor_if_cupy(values)
            return tf.RaggedTensor.from_row_splits(
                values=values, row_splits=[0, array.__len__()]
            )

        else:
            flat_values, nested_row_splits = _recursive_call(array, ())
            return tf.RaggedTensor.from_nested_row_splits(
                flat_values, nested_row_splits
            )


def _find_innermost_content(array):
    if isinstance(array, ak.contents.numpyarray.NumpyArray):
        return array
    else:
        return _find_innermost_content(array.content)


def _convert_to_tensor_if_cupy(array):
    if isinstance(array, np.ndarray):
        return array
    else:
        # converts cupy directly to tensor,
        # since `tf.RaggedTensor.from_nested_row_splits` can not work with Cupy arrays
        import tensorflow as tf

        return tf.experimental.dlpack.from_dlpack(array.toDlpack())


def _recursive_call(layout, offsets_arr):
    try:
        # change all the possible layout types to ListOffsetArray
        if isinstance(layout, ak.contents.listarray.ListArray):
            layout = layout.to_ListOffsetArray64()
        elif isinstance(layout, ak.contents.regulararray.RegularArray):
            layout = layout.to_ListOffsetArray64()
        elif not isinstance(
            layout,
            (
                ak.contents.listoffsetarray.ListOffsetArray,
                ak.contents.numpyarray.NumpyArray,
            ),
        ):
            raise TypeError(
                "Only arrays containing variable-length lists (var *) or"
                " regular-length lists (# *) of numbers can be converted into a TensorFlow RaggedTensor"
            )

        # recursively gather all of the offsets of an array
        offset = layout.offsets.data
        offset = _convert_to_tensor_if_cupy(offset)
        offsets_arr += (offset,)

    except AttributeError:
        # at the last iteration form a ragged tensor from the
        # accumulated offsets and flattened values of the array
        data = layout.data
        data = _convert_to_tensor_if_cupy(data)
        return data, offsets_arr
    return _recursive_call(layout.content, offsets_arr)



from __future__ import annotations

import numpy as np

import awkward as ak


def test_record_tuple():
    form = ak.forms.from_dict(
        {
            "class": "RecordArray",
            "fields": None,
            "contents": [
                {"class": "NumpyArray", "primitive": "int64", "form_key": "('0',)"},
                {"class": "NumpyArray", "primitive": "int64", "form_key": "('1',)"},
            ],
            "form_key": "()",
        }
    )
    array = ak.Array([(1, 2)])
    assert array.layout.form_with_key_path() == form


def test_record_dict():
    form = ak.forms.from_dict(
        {
            "class": "RecordArray",
            "fields": ["foo", "bar"],
            "contents": [
                {"class": "NumpyArray", "primitive": "int64", "form_key": "('foo',)"},
                {"class": "NumpyArray", "primitive": "int64", "form_key": "('bar',)"},
            ],
            "form_key": "()",
        }
    )
    array = ak.Array({"foo": [1], "bar": [2]})
    assert array.layout.form_with_key_path() == form


def test_numpy():
    form = ak.forms.from_dict(
        {"class": "NumpyArray", "primitive": "int64", "form_key": "()"}
    )
    array = ak.Array([1, 2, 3])
    assert array.layout.form_with_key_path() == form


def test_listoffset():
    form = ak.forms.from_dict(
        {
            "class": "ListOffsetArray",
            "offsets": "i64",
            "content": {
                "class": "NumpyArray",
                "primitive": "int64",
                "form_key": "(None,)",
            },
            "form_key": "()",
        }
    )
    array = ak.Array([[1, 2], [3]])
    assert array.layout.form_with_key_path() == form


def test_empty():
    form = ak.forms.from_dict({"class": "EmptyArray", "form_key": "()"})
    array = ak.Array([])
    assert array.layout.form_with_key_path() == form


def test_bitmasked():
    form = ak.forms.from_dict(
        {
            "class": "BitMaskedArray",
            "mask": "u8",
            "valid_when": True,
            "lsb_order": False,
            "content": {
                "class": "NumpyArray",
                "primitive": "float64",
                "form_key": "(None,)",
            },
            "form_key": "()",
        }
    )
    content = ak.Array([0.0, 1.1, 2.2, 3.3, 4.4]).layout
    mask = ak.index.IndexU8(
        np.packbits(np.array([False, True, True, False, False], dtype=np.int8))
    )
    bitmaskedarray = ak.contents.BitMaskedArray(mask, content, True, 5, False)
    assert bitmaskedarray.form_with_key_path() == form


def test_bytemasked():
    form = ak.forms.from_dict(
        {
            "class": "ByteMaskedArray",
            "mask": "i8",
            "valid_when": True,
            "content": {
                "class": "NumpyArray",
                "primitive": "float64",
                "form_key": "(None,)",
            },
            "form_key": "()",
        }
    )
    content = ak.Array([0.0, 1.1, 2.2, 3.3, 4.4]).layout
    mask = ak.index.Index8(np.array([False, True, True, False, False], dtype=np.int8))
    bytemaskedarray = ak.contents.ByteMaskedArray(mask, content, True)
    assert bytemaskedarray.form_with_key_path() == form


def test_indexedarray():
    form = ak.forms.from_dict(
        {
            "class": "IndexedArray",
            "index": "i64",
            "content": {
                "class": "NumpyArray",
                "primitive": "float64",
                "form_key": "(None,)",
            },
            "form_key": "()",
        }
    )

    content = ak.Array([0.0, 1.1, 2.2, 3.3, 4.4]).layout
    index = ak.index.Index64(np.array([3, 1, 1, 4, 2], dtype=np.int64))
    indexedarray = ak.contents.IndexedArray(index, content)
    assert indexedarray.form_with_key_path() == form


def test_indexedoptionarray():
    form = ak.forms.from_dict(
        {
            "class": "IndexedOptionArray",
            "index": "i64",
            "content": {
                "class": "NumpyArray",
                "primitive": "float64",
                "form_key": "(None,)",
            },
            "form_key": "()",
        }
    )
    content = ak.Array(
        np.array([0.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9])
    ).layout
    index = ak.index.Index64(np.array([2, -1, 4, 0, 8], dtype=np.int64))
    layout = ak.Array(ak.contents.IndexedOptionArray(index, content)).layout
    assert layout.form_with_key_path() == form


def test_listarray():
    form = ak.forms.from_dict(
        {
            "class": "ListArray",
            "starts": "i64",
            "stops": "i64",
            "content": {
                "class": "NumpyArray",
                "primitive": "float64",
                "form_key": "(None,)",
            },
            "form_key": "()",
        }
    )
    content = ak.contents.NumpyArray(
        np.array([1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9])
    )
    starts = ak.index.Index64(np.array([0, 3, 3, 5, 6]))
    stops = ak.index.Index64(np.array([3, 3, 5, 6, 9]))
    layout = ak.contents.ListArray(starts, stops, content)
    assert layout.form_with_key_path() == form


def test_regulararray():
    form = ak.forms.from_dict(
        {
            "class": "RegularArray",
            "size": 3,
            "content": {
                "class": "NumpyArray",
                "primitive": "int64",
                "form_key": "(None,)",
            },
            "form_key": "()",
        }
    )
    content = ak.Array([1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12]).layout
    layout = ak.contents.RegularArray(content, 3, zeros_length=0)
    assert layout.form_with_key_path() == form


def test_unionarray():
    form = ak.forms.from_dict(
        {
            "class": "UnionArray",
            "tags": "i8",
            "index": "i64",
            "contents": [
                {
                    "class": "ListOffsetArray",
                    "offsets": "i64",
                    "content": {
                        "class": "NumpyArray",
                        "primitive": "float64",
                        "form_key": "(0, None)",
                    },
                    "form_key": "(0,)",
                },
                {
                    "class": "ListOffsetArray",
                    "offsets": "i64",
                    "content": {
                        "class": "ListOffsetArray",
                        "offsets": "i64",
                        "content": {
                            "class": "NumpyArray",
                            "primitive": "uint8",
                            "parameters": {"__array__": "char"},
                            "form_key": "(1, None, None)",
                        },
                        "parameters": {"__array__": "string"},
                        "form_key": "(1, None)",
                    },
                    "form_key": "(1,)",
                },
            ],
            "form_key": "()",
        }
    )
    content1 = ak.operations.from_iter([[], [1.1], [2.2, 2.2]], highlevel=False)
    content2 = ak.operations.from_iter([["two", "two"], ["one"], []], highlevel=False)
    tags = ak.index.Index8(np.array([0, 1, 0, 1, 0, 1], dtype=np.int8))
    index = ak.index.Index64(np.array([0, 0, 1, 1, 2, 2], dtype=np.int64))
    layout = ak.contents.UnionArray(tags, index, [content1, content2])
    assert layout.form_with_key_path() == form


def test_unmaskedarray():
    form = ak.forms.from_dict(
        {
            "class": "UnmaskedArray",
            "content": {
                "class": "NumpyArray",
                "primitive": "int64",
                "form_key": "(None,)",
            },
            "form_key": "()",
        }
    )

    content = ak.Array([1, 2, 3, 4, 5]).layout
    layout = ak.contents.UnmaskedArray(content)
    assert layout.form_with_key_path() == form



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations


def test_prettyprint_rename():
    import awkward._prettyprint as deprecated_prettyprint
    import awkward.prettyprint as new_prettyprint

    assert new_prettyprint.Formatter is deprecated_prettyprint.Formatter
    assert new_prettyprint.FormatterOptions is deprecated_prettyprint.FormatterOptions
    assert new_prettyprint.FormatterType is deprecated_prettyprint.FormatterType
    assert new_prettyprint.PlaceholderValue is deprecated_prettyprint.PlaceholderValue
    assert new_prettyprint.alternate is deprecated_prettyprint.alternate
    assert new_prettyprint.custom_str is deprecated_prettyprint.custom_str
    assert new_prettyprint.get_at is deprecated_prettyprint.get_at
    assert new_prettyprint.get_field is deprecated_prettyprint.get_field
    assert new_prettyprint.half is deprecated_prettyprint.half
    assert new_prettyprint.is_identifier is deprecated_prettyprint.is_identifier
    assert new_prettyprint.valuestr is deprecated_prettyprint.valuestr
    assert new_prettyprint.valuestr_horiz is deprecated_prettyprint.valuestr_horiz


def test_remove_structure_rename():
    from awkward._do import remove_structure as deprecated_remove_structure
    from awkward.contents.remove_structure import (
        remove_structure as new_remove_structure,
    )

    assert new_remove_structure is deprecated_remove_structure



# AUTO GENERATED ON 2026-02-08 AT 18:23:02
# DO NOT EDIT BY HAND!
#
# To regenerate file, run
#
#     python dev/generate-tests.py
#

# fmt: off

import cupy
import cupy.testing as cpt
import numpy as np
import pytest

import awkward as ak
import awkward._connect.cuda as ak_cu
from awkward._backends.cupy import CupyBackend

cupy_backend = CupyBackend.instance()

def test_cudaawkward_ListArray64_getitem_next_range_counts_64_1():
    total = cupy.array([123], dtype=cupy.int64)
    fromoffsets = cupy.array([1, 1, 1, 1, 1, 1, 1, 1, 1, 1], dtype=cupy.int64)
    lenstarts = 3
    funcC = cupy_backend['awkward_ListArray_getitem_next_range_counts', cupy.int64, cupy.int64]
    funcC(total, fromoffsets, lenstarts)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_total = [0]
    cpt.assert_allclose(total[:len(pytest_total)], cupy.array(pytest_total))

def test_cudaawkward_ListArray64_getitem_next_range_counts_64_2():
    total = cupy.array([123], dtype=cupy.int64)
    fromoffsets = cupy.array([2, 3, 3, 4, 5, 5, 5, 5, 5, 6, 7, 8, 10, 11], dtype=cupy.int64)
    lenstarts = 3
    funcC = cupy_backend['awkward_ListArray_getitem_next_range_counts', cupy.int64, cupy.int64]
    funcC(total, fromoffsets, lenstarts)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_total = [2]
    cpt.assert_allclose(total[:len(pytest_total)], cupy.array(pytest_total))

def test_cudaawkward_ListArray64_getitem_next_range_counts_64_3():
    total = cupy.array([123], dtype=cupy.int64)
    fromoffsets = cupy.array([2, 1, 0, 1, 2, 0, 1, 2, 2, 2, 1, 2, 1, 0, 0, 0, 0], dtype=cupy.int64)
    lenstarts = 3
    funcC = cupy_backend['awkward_ListArray_getitem_next_range_counts', cupy.int64, cupy.int64]
    funcC(total, fromoffsets, lenstarts)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_total = [-1]
    cpt.assert_allclose(total[:len(pytest_total)], cupy.array(pytest_total))

def test_cudaawkward_ListArray64_getitem_next_range_counts_64_4():
    total = cupy.array([123], dtype=cupy.int64)
    fromoffsets = cupy.array([1, 0, 2, 3, 1, 2, 0, 0, 1, 1, 2, 3, 1, 2, 3, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1], dtype=cupy.int64)
    lenstarts = 3
    funcC = cupy_backend['awkward_ListArray_getitem_next_range_counts', cupy.int64, cupy.int64]
    funcC(total, fromoffsets, lenstarts)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_total = [2]
    cpt.assert_allclose(total[:len(pytest_total)], cupy.array(pytest_total))

def test_cudaawkward_ListArray64_getitem_next_range_counts_64_5():
    total = cupy.array([123], dtype=cupy.int64)
    fromoffsets = cupy.array([0, 0, 0, 0, 0, 0, 0, 0], dtype=cupy.int64)
    lenstarts = 3
    funcC = cupy_backend['awkward_ListArray_getitem_next_range_counts', cupy.int64, cupy.int64]
    funcC(total, fromoffsets, lenstarts)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_total = [0]
    cpt.assert_allclose(total[:len(pytest_total)], cupy.array(pytest_total))




# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE
#
from __future__ import annotations

import numpy as np

from awkward._nplikes.numpy import Numpy
from awkward._nplikes.virtual import VirtualNDArray


def test():
    numpy_like = Numpy.instance()
    v1 = VirtualNDArray(
        numpy_like,
        shape=(10,),
        dtype=np.dtype(np.float32),
        generator=lambda: np.arange(10, dtype=np.float32),
    )
    v2 = v1.copy()
    np.testing.assert_array_equal(v1.materialize(), np.arange(10, dtype=np.float32))
    np.testing.assert_array_equal(v2.materialize(), np.arange(10, dtype=np.float32))

    v1 = VirtualNDArray(
        numpy_like,
        shape=(10,),
        dtype=np.dtype(np.float32),
        generator=lambda: np.arange(10, dtype=np.float32),
    )
    v2 = v1.copy()
    np.testing.assert_array_equal(v2.materialize(), np.arange(10, dtype=np.float32))
    np.testing.assert_array_equal(v1.materialize(), np.arange(10, dtype=np.float32))



from __future__ import annotations

import cupy as cp
import cupy.testing as cpt
import numpy as np
import pytest

import awkward as ak

to_list = ak.operations.to_list


@pytest.fixture(scope="function", autouse=True)
def cleanup_cuda():
    yield
    try:
        cp.cuda.Device().synchronize()  # wait for all kernels
    except cp.cuda.runtime.CUDARuntimeError as e:
        print("GPU error during sync:", e)
    cp._default_memory_pool.free_all_blocks()


def test_0652_tests_of_complex_numbers_reducers():
    array = [[1 + 1j, 2 + 2j], [], [3 + 3j]]
    cuda_array = ak.to_backend(array, "cuda")

    assert ak.operations.sum(ak.operations.from_iter(cuda_array)) == 6 + 6j

    assert ak.operations.sum(ak.operations.from_iter(cuda_array), axis=1).to_list() == [
        3 + 3j,
        0 + 0j,
        3 + 3j,
    ]
    del cuda_array


def test_0395_astype_complex():
    content_float64 = ak.contents.NumpyArray(
        np.array([0.25, 0.5, 3.5, 4.5, 5.5], dtype=np.float64)
    )
    cuda_content_float64 = ak.to_backend(content_float64, "cuda", highlevel=False)

    array_float64 = ak.contents.UnmaskedArray(content_float64)
    cuda_array_float64 = ak.to_backend(array_float64, "cuda", highlevel=False)

    assert to_list(cuda_array_float64) == [0.25, 0.5, 3.5, 4.5, 5.5]
    assert str(ak.operations.type(cuda_content_float64)) == "5 * float64"
    assert (
        str(ak.operations.type(ak.highlevel.Array(cuda_content_float64)))
        == "5 * float64"
    )
    assert str(ak.operations.type(cuda_array_float64)) == "5 * ?float64"
    assert (
        str(ak.operations.type(ak.highlevel.Array(cuda_array_float64)))
        == "5 * ?float64"
    )
    assert str(cuda_content_float64.form.type) == "float64"
    assert str(cuda_array_float64.form.type) == "?float64"

    del cuda_content_float64, cuda_array_float64

    content_complex64 = ak.operations.values_astype(
        content_float64, "complex64", highlevel=False
    )
    cuda_content_complex64 = ak.to_backend(content_complex64, "cuda", highlevel=False)

    array_complex64 = ak.contents.UnmaskedArray(content_complex64)
    cuda_array_complex64 = ak.to_backend(array_complex64, "cuda", highlevel=False)

    assert to_list(cuda_content_complex64) == [
        (0.25 + 0j),
        (0.5 + 0j),
        (3.5 + 0j),
        (4.5 + 0j),
        (5.5 + 0j),
    ]
    assert to_list(cuda_array_complex64) == [
        (0.25 + 0.0j),
        (0.5 + 0.0j),
        (3.5 + 0.0j),
        (4.5 + 0.0j),
        (5.5 + 0.0j),
    ]
    assert str(ak.operations.type(cuda_content_complex64)) == "5 * complex64"
    assert (
        str(ak.operations.type(ak.highlevel.Array(cuda_content_complex64)))
        == "5 * complex64"
    )
    assert str(ak.operations.type(cuda_array_complex64)) == "5 * ?complex64"
    assert (
        str(ak.operations.type(ak.highlevel.Array(cuda_array_complex64)))
        == "5 * ?complex64"
    )
    assert str(cuda_content_complex64.form.type) == "complex64"
    assert str(cuda_array_complex64.form.type) == "?complex64"

    content = ak.contents.NumpyArray(
        np.array([1, (2.2 + 0.1j), 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9])
    )
    cuda_content = ak.to_backend(content, "cuda")

    assert to_list(cuda_content) == [
        (1 + 0j),
        (2.2 + 0.1j),
        (3.3 + 0j),
        (4.4 + 0j),
        (5.5 + 0j),
        (6.6 + 0j),
        (7.7 + 0j),
        (8.8 + 0j),
        (9.9 + 0j),
    ]

    assert ak.sum(cuda_content_complex64) == (14.25 + 0j)
    assert ak.min(content_complex64, highlevel=False) == (0.25 + 0j)
    assert ak.max(content_complex64, highlevel=False) == (5.5 + 0j)
    del cuda_content_complex64, cuda_array_complex64
    del content


def test_0395_count_complex():
    content2 = ak.contents.NumpyArray(
        np.array([(1.1 + 0.1j), 2.2, 3.3, 0.0, 2.2, 0.0, 0.0, 2.2, 0.0, 4.4])
    )
    offsets3 = ak.index.Index64(np.array([0, 3, 6, 10], dtype=np.int64))
    depth1 = ak.contents.ListOffsetArray(offsets3, content2)
    depth1 = ak.to_backend(depth1, "cuda", highlevel=False)

    assert to_list(depth1) == [
        [(1.1 + 0.1j), (2.2 + 0j), (3.3 + 0j)],
        [0j, (2.2 + 0j), 0j],
        [0j, (2.2 + 0j), 0j, (4.4 + 0j)],
    ]

    assert to_list(ak.count(depth1, -1, highlevel=False)) == [3, 3, 4]
    assert to_list(ak.count(depth1, 1, highlevel=False)) == [3, 3, 4]

    del depth1


def test_0395_mask_complex():
    content = ak.contents.NumpyArray(
        np.array([(1.1 + 0.1j), 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9])
    )
    offsets = ak.index.Index64(np.array([0, 3, 3, 5, 6, 6, 6, 9], dtype=np.int64))
    array = ak.contents.ListOffsetArray(offsets, content)
    array = ak.to_backend(array, "cuda", highlevel=False)

    assert to_list(ak.min(array, axis=-1, mask_identity=False, highlevel=False)) == [
        (1.1 + 0.1j),
        (np.inf + 0j),
        (4.4 + 0j),
        (6.6 + 0j),
        (np.inf + 0j),
        (np.inf + 0j),
        (7.7 + 0j),
    ]
    assert to_list(ak.min(array, axis=-1, mask_identity=True, highlevel=False)) == [
        (1.1 + 0.1j),
        None,
        (4.4 + 0j),
        (6.6 + 0j),
        None,
        None,
        (7.7 + 0j),
    ]
    del array


def test_0395_count_min_complex():
    content2 = ak.contents.NumpyArray(
        np.array([(1.1 + 0.1j), 2.2, 3.3, 0.0, 2.2, 0.0, 0.0, 2.2, 0.0, 4.4])
    )
    offsets3 = ak.index.Index64(np.array([0, 3, 6, 10], dtype=np.int64))
    depth1 = ak.contents.ListOffsetArray(offsets3, content2)
    depth1 = ak.to_backend(depth1, "cuda", highlevel=False)

    assert to_list(depth1) == [
        [(1.1 + 0.1j), (2.2 + 0j), (3.3 + 0j)],
        [0j, (2.2 + 0j), 0j],
        [0j, (2.2 + 0j), 0j, (4.4 + 0j)],
    ]

    assert to_list(ak.min(depth1, -1, highlevel=False)) == [(1.1 + 0.1j), 0j, 0j]
    assert to_list(ak.min(depth1, 1, highlevel=False)) == [(1.1 + 0.1j), 0j, 0j]

    content2 = ak.contents.NumpyArray(
        np.array([True, True, True, False, True, False, False, True, False, True])
    )
    offsets3 = ak.index.Index64(np.array([0, 3, 6, 10], dtype=np.int64))
    depth1 = ak.contents.ListOffsetArray(offsets3, content2)
    assert to_list(depth1) == [
        [True, True, True],
        [False, True, False],
        [False, True, False, True],
    ]

    assert to_list(ak.min(depth1, -1, highlevel=False)) == [True, False, False]
    assert to_list(ak.min(depth1, 1, highlevel=False)) == [True, False, False]
    del depth1


def test_0395_count_max_complex():
    content2 = ak.contents.NumpyArray(
        np.array([(1.1 + 0.1j), 2.2, 3.3, 0.0, 2.2, 0.0, 0.0, 2.2, 0.0, 4.4])
    )
    offsets3 = ak.index.Index64(np.array([0, 3, 6, 10], dtype=np.int64))
    depth1 = ak.contents.ListOffsetArray(offsets3, content2)
    depth1 = ak.to_backend(depth1, "cuda", highlevel=False)

    assert to_list(depth1) == [
        [(1.1 + 0.1j), (2.2 + 0j), (3.3 + 0j)],
        [0j, (2.2 + 0j), 0j],
        [0j, (2.2 + 0j), 0j, (4.4 + 0j)],
    ]

    assert to_list(ak.max(depth1, -1, highlevel=False)) == [3.3, 2.2, 4.4]
    assert to_list(ak.max(depth1, 1, highlevel=False)) == [3.3, 2.2, 4.4]

    content2 = ak.contents.NumpyArray(
        np.array([False, True, True, False, True, False, False, False, False, False])
    )
    offsets3 = ak.index.Index64(np.array([0, 3, 6, 10], dtype=np.int64))
    depth1 = ak.contents.ListOffsetArray(offsets3, content2)
    assert to_list(depth1) == [
        [False, True, True],
        [False, True, False],
        [False, False, False, False],
    ]

    assert to_list(ak.max(depth1, -1, highlevel=False)) == [True, True, False]
    assert to_list(ak.max(depth1, 1, highlevel=False)) == [True, True, False]
    del depth1


def test_0395_count_nonzero_complex():
    content2 = ak.contents.NumpyArray(
        np.array([(1.1 + 0.1j), 2.2, 3.3, 0.0, 2.2, 0.0, 0.0, 2.2, 0.0, 4.4])
    )
    offsets3 = ak.index.Index64(np.array([0, 3, 6, 10], dtype=np.int64))
    depth1 = ak.contents.ListOffsetArray(offsets3, content2)
    depth1 = ak.to_backend(depth1, "cuda", highlevel=False)

    assert to_list(depth1) == [
        [(1.1 + 0.1j), (2.2 + 0j), (3.3 + 0j)],
        [0j, (2.2 + 0j), 0j],
        [0j, (2.2 + 0j), 0j, (4.4 + 0j)],
    ]

    assert to_list(ak.count_nonzero(depth1, -1, highlevel=False)) == [3, 1, 2]
    assert to_list(ak.count_nonzero(depth1, 1, highlevel=False)) == [3, 1, 2]

    del depth1


def test_0652_reducers():
    array = ak.operations.from_iter([[1 + 1j, 2 + 2j], [], [3 + 3j]])
    array = ak.to_backend(array, "cuda", highlevel=False)

    assert ak.operations.sum(array) == 6 + 6j
    assert ak.operations.prod(array) == -12 + 12j

    assert ak.operations.sum(array, axis=1).to_list() == [
        3 + 3j,
        0 + 0j,
        3 + 3j,
    ]
    assert ak.operations.prod(array, axis=1).to_list() == [
        0 + 4j,
        1 + 0j,
        3 + 3j,
    ]

    assert ak.operations.count(array, axis=1).to_list() == [2, 0, 1]
    assert ak.operations.count_nonzero(array, axis=1).to_list() == [2, 0, 1]
    assert ak.operations.any(array, axis=1).to_list() == [True, False, True]
    assert ak.operations.all(array, axis=1).to_list() == [True, True, True]

    array = ak.operations.from_iter([[1 + 1j, 2 + 2j, 0 + 0j], [], [3 + 3j]])
    array = ak.to_backend(array, "cuda", highlevel=False)

    assert ak.operations.any(array, axis=1).to_list() == [True, False, True]
    assert ak.operations.all(array, axis=1).to_list() == [False, True, True]
    del array


def test_0652_minmax():
    array = ak.operations.from_iter([[1 + 5j, 2 + 4j], [], [3 + 3j]])
    array = ak.to_backend(array, "cuda", highlevel=False)

    assert ak.operations.min(array) == 1 + 5j
    assert ak.operations.max(array) == 3 + 3j

    assert ak.operations.min(array, axis=1).to_list() == [
        1 + 5j,
        None,
        3 + 3j,
    ]
    assert ak.operations.max(array, axis=1).to_list() == [
        2 + 4j,
        None,
        3 + 3j,
    ]
    del array


def test_block_boundary_sum_complex():
    rng = np.random.default_rng(seed=42)
    array = rng.integers(6000, size=6000)
    complex_array = np.vectorize(complex)(
        array[0 : len(array) : 2], array[1 : len(array) : 2]
    )
    content = ak.contents.NumpyArray(complex_array)
    cuda_content = ak.to_backend(content, "cuda", highlevel=False)
    assert ak.sum(cuda_content, -1, highlevel=False) == ak.sum(
        content, -1, highlevel=False
    )

    offsets = ak.index.Index64(np.array([0, 1, 2998, 3000], dtype=np.int64))
    depth1 = ak.contents.ListOffsetArray(offsets, content)
    cuda_depth1 = ak.to_backend(depth1, "cuda", highlevel=False)
    assert to_list(ak.sum(cuda_depth1, -1, highlevel=False)) == to_list(
        ak.sum(depth1, -1, highlevel=False)
    )
    del cuda_content, cuda_depth1


def test_block_boundary_prod_complex1():
    complex_array = np.vectorize(complex)(np.full(1000, 0), np.full(1000, 1))
    content = ak.contents.NumpyArray(complex_array)
    cuda_content = ak.to_backend(content, "cuda", highlevel=False)
    assert ak.prod(cuda_content, -1, highlevel=False) == ak.prod(
        content, -1, highlevel=False
    )

    offsets = ak.index.Index64(np.array([0, 5, 996, 1000], dtype=np.int64))
    depth1 = ak.contents.ListOffsetArray(offsets, content)
    cuda_depth1 = ak.to_backend(depth1, "cuda", highlevel=False)
    assert to_list(ak.prod(cuda_depth1, -1, highlevel=False)) == to_list(
        ak.prod(depth1, -1, highlevel=False)
    )
    del cuda_content, cuda_depth1


def test_block_boundary_prod_complex2():
    complex_array = np.vectorize(complex)(np.full(1001, 0), np.full(1001, 1))
    content = ak.contents.NumpyArray(complex_array)
    cuda_content = ak.to_backend(content, "cuda", highlevel=False)
    assert ak.prod(cuda_content, -1, highlevel=False) == ak.prod(
        content, -1, highlevel=False
    )

    offsets = ak.index.Index64(np.array([0, 5, 996, 1001], dtype=np.int64))
    depth1 = ak.contents.ListOffsetArray(offsets, content)
    cuda_depth1 = ak.to_backend(depth1, "cuda", highlevel=False)
    assert to_list(ak.prod(cuda_depth1, -1, highlevel=False)) == to_list(
        ak.prod(depth1, -1, highlevel=False)
    )
    del cuda_content, cuda_depth1


def test_block_boundary_prod_complex3():
    complex_array = np.vectorize(complex)(np.full(1002, 0), np.full(1002, 1))
    content = ak.contents.NumpyArray(complex_array)
    cuda_content = ak.to_backend(content, "cuda", highlevel=False)
    assert ak.prod(cuda_content, -1, highlevel=False) == ak.prod(
        content, -1, highlevel=False
    )

    offsets = ak.index.Index64(np.array([0, 5, 999, 1002], dtype=np.int64))
    depth1 = ak.contents.ListOffsetArray(offsets, content)
    cuda_depth1 = ak.to_backend(depth1, "cuda", highlevel=False)
    assert to_list(ak.prod(cuda_depth1, -1, highlevel=False)) == to_list(
        ak.prod(depth1, -1, highlevel=False)
    )
    del cuda_content, cuda_depth1


def test_block_boundary_prod_complex4():
    complex_array = np.vectorize(complex)(np.full(1000, 0), np.full(1000, 1.01))
    content = ak.contents.NumpyArray(complex_array)
    cuda_content = ak.to_backend(content, "cuda", highlevel=False)
    cpt.assert_allclose(
        ak.prod(cuda_content, -1, highlevel=False),
        ak.prod(content, -1, highlevel=False),
    )

    offsets = ak.index.Index64(np.array([0, 5, 996, 1000], dtype=np.int64))
    depth1 = ak.contents.ListOffsetArray(offsets, content)
    cuda_depth1 = ak.to_backend(depth1, "cuda", highlevel=False)
    cpt.assert_allclose(
        to_list(ak.prod(cuda_depth1, -1, highlevel=False)),
        to_list(ak.prod(depth1, -1, highlevel=False)),
    )
    del cuda_content, cuda_depth1


def test_block_boundary_prod_complex5():
    complex_array = np.vectorize(complex)(np.full(1001, 0), np.full(1001, 1.01))
    content = ak.contents.NumpyArray(complex_array)
    cuda_content = ak.to_backend(content, "cuda", highlevel=False)
    cpt.assert_allclose(
        ak.prod(cuda_content, -1, highlevel=False),
        ak.prod(content, -1, highlevel=False),
    )

    offsets = ak.index.Index64(np.array([0, 5, 996, 1001], dtype=np.int64))
    depth1 = ak.contents.ListOffsetArray(offsets, content)
    cuda_depth1 = ak.to_backend(depth1, "cuda", highlevel=False)
    cpt.assert_allclose(
        to_list(ak.prod(cuda_depth1, -1, highlevel=False)),
        to_list(ak.prod(depth1, -1, highlevel=False)),
    )
    del cuda_content, cuda_depth1


def test_block_boundary_prod_complex6():
    complex_array = np.vectorize(complex)(np.full(1002, 0), np.full(1002, 1.01))
    content = ak.contents.NumpyArray(complex_array)
    cuda_content = ak.to_backend(content, "cuda", highlevel=False)
    cpt.assert_allclose(
        ak.prod(cuda_content, -1, highlevel=False),
        ak.prod(content, -1, highlevel=False),
    )

    offsets = ak.index.Index64(np.array([0, 5, 998, 1002], dtype=np.int64))
    depth1 = ak.contents.ListOffsetArray(offsets, content)
    cuda_depth1 = ak.to_backend(depth1, "cuda", highlevel=False)
    cpt.assert_allclose(
        to_list(ak.prod(cuda_depth1, -1, highlevel=False)),
        to_list(ak.prod(depth1, -1, highlevel=False)),
    )
    del cuda_content, cuda_depth1


def test_block_boundary_prod_complex7():
    complex_array = np.vectorize(complex)(np.full(1000, 0), np.full(1000, 0.99))
    content = ak.contents.NumpyArray(complex_array)
    cuda_content = ak.to_backend(content, "cuda", highlevel=False)
    cpt.assert_allclose(
        ak.prod(cuda_content, -1, highlevel=False),
        ak.prod(content, -1, highlevel=False),
    )

    offsets = ak.index.Index64(np.array([0, 5, 996, 1000], dtype=np.int64))
    depth1 = ak.contents.ListOffsetArray(offsets, content)
    cuda_depth1 = ak.to_backend(depth1, "cuda", highlevel=False)
    cpt.assert_allclose(
        to_list(ak.prod(cuda_depth1, -1, highlevel=False)),
        to_list(ak.prod(depth1, -1, highlevel=False)),
    )
    del cuda_content, cuda_depth1


def test_block_boundary_prod_complex8():
    complex_array = np.vectorize(complex)(np.full(1001, 0), np.full(1001, 0.99))
    content = ak.contents.NumpyArray(complex_array)
    cuda_content = ak.to_backend(content, "cuda", highlevel=False)
    cpt.assert_allclose(
        ak.prod(cuda_content, -1, highlevel=False),
        ak.prod(content, -1, highlevel=False),
    )

    offsets = ak.index.Index64(np.array([0, 5, 996, 1001], dtype=np.int64))
    depth1 = ak.contents.ListOffsetArray(offsets, content)
    cuda_depth1 = ak.to_backend(depth1, "cuda", highlevel=False)
    cpt.assert_allclose(
        to_list(ak.prod(cuda_depth1, -1, highlevel=False)),
        to_list(ak.prod(depth1, -1, highlevel=False)),
    )
    del cuda_content, cuda_depth1


def test_block_boundary_prod_complex9():
    complex_array = np.vectorize(complex)(np.full(1002, 0), np.full(1002, 0.99))
    content = ak.contents.NumpyArray(complex_array)
    cuda_content = ak.to_backend(content, "cuda", highlevel=False)
    cpt.assert_allclose(
        ak.prod(cuda_content, -1, highlevel=False),
        ak.prod(content, -1, highlevel=False),
    )

    offsets = ak.index.Index64(np.array([0, 5, 999, 1002], dtype=np.int64))
    depth1 = ak.contents.ListOffsetArray(offsets, content)
    cuda_depth1 = ak.to_backend(depth1, "cuda", highlevel=False)
    cpt.assert_allclose(
        to_list(ak.prod(cuda_depth1, -1, highlevel=False)),
        to_list(ak.prod(depth1, -1, highlevel=False)),
    )
    del cuda_content, cuda_depth1


def test_block_boundary_prod_complex10():
    complex_array = np.vectorize(complex)(np.full(1000, 0), np.full(1000, 1.1))
    content = ak.contents.NumpyArray(complex_array)
    cuda_content = ak.to_backend(content, "cuda", highlevel=False)
    cpt.assert_allclose(
        ak.prod(cuda_content, -1, highlevel=False),
        ak.prod(content, -1, highlevel=False),
    )

    offsets = ak.index.Index64(np.array([0, 5, 996, 1000], dtype=np.int64))
    depth1 = ak.contents.ListOffsetArray(offsets, content)
    cuda_depth1 = ak.to_backend(depth1, "cuda", highlevel=False)
    cpt.assert_allclose(
        to_list(ak.prod(cuda_depth1, -1, highlevel=False)),
        to_list(ak.prod(depth1, -1, highlevel=False)),
    )
    del cuda_content, cuda_depth1


def test_block_boundary_prod_complex11():
    complex_array = np.vectorize(complex)(np.full(1001, 0), np.full(1001, 1.1))
    content = ak.contents.NumpyArray(complex_array)
    cuda_content = ak.to_backend(content, "cuda", highlevel=False)
    cpt.assert_allclose(
        ak.prod(cuda_content, -1, highlevel=False),
        ak.prod(content, -1, highlevel=False),
    )

    offsets = ak.index.Index64(np.array([0, 5, 996, 1001], dtype=np.int64))
    depth1 = ak.contents.ListOffsetArray(offsets, content)
    cuda_depth1 = ak.to_backend(depth1, "cuda", highlevel=False)
    cpt.assert_allclose(
        to_list(ak.prod(cuda_depth1, -1, highlevel=False)),
        to_list(ak.prod(depth1, -1, highlevel=False)),
    )
    del cuda_content, cuda_depth1


def test_block_boundary_prod_complex12():
    complex_array = np.vectorize(complex)(np.full(1002, 0), np.full(1002, 1.1))
    content = ak.contents.NumpyArray(complex_array)
    cuda_content = ak.to_backend(content, "cuda", highlevel=False)
    cpt.assert_allclose(
        ak.prod(cuda_content, -1, highlevel=False),
        ak.prod(content, -1, highlevel=False),
    )

    offsets = ak.index.Index64(np.array([0, 5, 996, 1002], dtype=np.int64))
    depth1 = ak.contents.ListOffsetArray(offsets, content)
    cuda_depth1 = ak.to_backend(depth1, "cuda", highlevel=False)
    cpt.assert_allclose(
        to_list(ak.prod(cuda_depth1, -1, highlevel=False)),
        to_list(ak.prod(depth1, -1, highlevel=False)),
    )
    del cuda_content, cuda_depth1


def test_block_boundary_prod_complex13():
    rng = np.random.default_rng(seed=42)
    array = rng.integers(50, size=1000)
    complex_array = np.vectorize(complex)(
        array[0 : len(array) : 2], array[1 : len(array) : 2]
    )
    content = ak.contents.NumpyArray(complex_array)
    cuda_content = ak.to_backend(content, "cuda", highlevel=False)
    cpt.assert_allclose(
        ak.prod(cuda_content, -1, highlevel=False),
        ak.prod(content, -1, highlevel=False),
    )

    offsets = ak.index.Index64(np.array([0, 5, 996, 1000], dtype=np.int64))
    depth1 = ak.contents.ListOffsetArray(offsets, content)
    cuda_depth1 = ak.to_backend(depth1, "cuda", highlevel=False)
    cpt.assert_allclose(
        to_list(ak.prod(cuda_depth1, -1, highlevel=False)),
        to_list(ak.prod(depth1, -1, highlevel=False)),
    )
    del cuda_content, cuda_depth1


def test_block_boundary_any_complex():
    rng = np.random.default_rng(seed=42)
    array = rng.integers(6000, size=6000)
    complex_array = np.vectorize(complex)(
        array[0 : len(array) : 2], array[1 : len(array) : 2]
    )
    content = ak.contents.NumpyArray(complex_array)
    cuda_content = ak.to_backend(content, "cuda", highlevel=False)
    assert ak.any(cuda_content, -1, highlevel=False) == ak.any(
        content, -1, highlevel=False
    )

    offsets = ak.index.Index64(np.array([0, 1, 2998, 3000], dtype=np.int64))
    depth1 = ak.contents.ListOffsetArray(offsets, content)
    cuda_depth1 = ak.to_backend(depth1, "cuda", highlevel=False)
    assert to_list(ak.any(cuda_depth1, -1, highlevel=False)) == to_list(
        ak.any(depth1, -1, highlevel=False)
    )
    del cuda_content, cuda_depth1


def test_block_boundary_all_complex():
    rng = np.random.default_rng(seed=42)
    array = rng.integers(6000, size=6000)
    complex_array = np.vectorize(complex)(
        array[0 : len(array) : 2], array[1 : len(array) : 2]
    )
    content = ak.contents.NumpyArray(complex_array)
    cuda_content = ak.to_backend(content, "cuda", highlevel=False)
    assert ak.all(cuda_content, -1, highlevel=False) == ak.all(
        content, -1, highlevel=False
    )

    offsets = ak.index.Index64(np.array([0, 1, 2998, 3000], dtype=np.int64))
    depth1 = ak.contents.ListOffsetArray(offsets, content)
    cuda_depth1 = ak.to_backend(depth1, "cuda", highlevel=False)
    assert to_list(ak.all(cuda_depth1, -1, highlevel=False)) == to_list(
        ak.all(depth1, -1, highlevel=False)
    )
    del cuda_content, cuda_depth1


def test_block_boundary_min_complex1():
    rng = np.random.default_rng(seed=42)
    array = rng.integers(5, size=6000)
    complex_array = np.vectorize(complex)(
        array[0 : len(array) : 2], array[1 : len(array) : 2]
    )
    content = ak.contents.NumpyArray(complex_array)
    cuda_content = ak.to_backend(content, "cuda", highlevel=False)
    assert ak.min(cuda_content, -1, highlevel=False) == ak.min(
        content, -1, highlevel=False
    )

    offsets = ak.index.Index64(np.array([0, 1, 2998, 3000], dtype=np.int64))
    depth1 = ak.contents.ListOffsetArray(offsets, content)
    cuda_depth1 = ak.to_backend(depth1, "cuda", highlevel=False)
    assert to_list(ak.min(cuda_depth1, -1, highlevel=False)) == to_list(
        ak.min(depth1, -1, highlevel=False)
    )
    del cuda_content, cuda_depth1


def test_block_boundary_min_complex2():
    rng = np.random.default_rng(seed=42)
    array = rng.integers(6000, size=6000)
    complex_array = np.vectorize(complex)(
        array[0 : len(array) : 2], array[1 : len(array) : 2]
    )
    content = ak.contents.NumpyArray(complex_array)
    cuda_content = ak.to_backend(content, "cuda", highlevel=False)
    assert ak.min(cuda_content, -1, highlevel=False) == ak.min(
        content, -1, highlevel=False
    )

    offsets = ak.index.Index64(np.array([0, 1, 2998, 3000], dtype=np.int64))
    depth1 = ak.contents.ListOffsetArray(offsets, content)
    cuda_depth1 = ak.to_backend(depth1, "cuda", highlevel=False)
    assert to_list(ak.min(cuda_depth1, -1, highlevel=False)) == to_list(
        ak.min(depth1, -1, highlevel=False)
    )
    del cuda_content, cuda_depth1


def test_block_boundary_max_complex1():
    rng = np.random.default_rng(seed=42)
    array = rng.integers(5, size=6000)
    complex_array = np.vectorize(complex)(
        array[0 : len(array) : 2], array[1 : len(array) : 2]
    )
    content = ak.contents.NumpyArray(complex_array)
    cuda_content = ak.to_backend(content, "cuda", highlevel=False)
    assert ak.max(cuda_content, -1, highlevel=False) == ak.max(
        content, -1, highlevel=False
    )

    offsets = ak.index.Index64(np.array([0, 1, 2998, 3000], dtype=np.int64))
    depth1 = ak.contents.ListOffsetArray(offsets, content)
    cuda_depth1 = ak.to_backend(depth1, "cuda", highlevel=False)
    assert to_list(ak.max(cuda_depth1, -1, highlevel=False)) == to_list(
        ak.max(depth1, -1, highlevel=False)
    )
    del cuda_content, cuda_depth1


def test_block_boundary_max_complex2():
    rng = np.random.default_rng(seed=42)
    array = rng.integers(6000, size=6000)
    complex_array = np.vectorize(complex)(
        array[0 : len(array) : 2], array[1 : len(array) : 2]
    )
    content = ak.contents.NumpyArray(complex_array)
    cuda_content = ak.to_backend(content, "cuda", highlevel=False)
    assert ak.max(cuda_content, -1, highlevel=False) == ak.max(
        content, -1, highlevel=False
    )

    offsets = ak.index.Index64(np.array([0, 1, 2998, 3000], dtype=np.int64))
    depth1 = ak.contents.ListOffsetArray(offsets, content)
    cuda_depth1 = ak.to_backend(depth1, "cuda", highlevel=False)
    assert to_list(ak.max(cuda_depth1, -1, highlevel=False)) == to_list(
        ak.max(depth1, -1, highlevel=False)
    )
    del cuda_content, cuda_depth1


def test_block_boundary_sum_bool_complex():
    rng = np.random.default_rng(seed=42)
    array = rng.integers(2, size=6000, dtype=np.bool_)
    complex_array = np.vectorize(complex)(
        array[0 : len(array) : 2], array[1 : len(array) : 2]
    )
    content = ak.contents.NumpyArray(complex_array)
    cuda_content = ak.to_backend(content, "cuda", highlevel=False)
    assert ak.sum(cuda_content, -1, highlevel=False) == ak.sum(
        content, -1, highlevel=False
    )

    offsets = ak.index.Index64(np.array([0, 1, 2998, 3000], dtype=np.int64))
    depth1 = ak.contents.ListOffsetArray(offsets, content)
    cuda_depth1 = ak.to_backend(depth1, "cuda", highlevel=False)
    assert to_list(ak.sum(cuda_depth1, -1, highlevel=False)) == to_list(
        ak.sum(depth1, -1, highlevel=False)
    )
    del cuda_content, cuda_depth1


def test_block_boundary_countnonzero_complex_1():
    rng = np.random.default_rng(seed=42)
    array = rng.integers(6000, size=6000)
    complex_array = np.vectorize(complex)(
        array[0 : len(array) : 2], array[1 : len(array) : 2]
    )
    content = ak.contents.NumpyArray(complex_array)
    cuda_content = ak.to_backend(content, "cuda", highlevel=False)
    assert ak.count_nonzero(cuda_content, -1, highlevel=False) == ak.count_nonzero(
        content, -1, highlevel=False
    )

    offsets = ak.index.Index64(np.array([0, 1, 2998, 3000], dtype=np.int64))
    depth1 = ak.contents.ListOffsetArray(offsets, content)
    cuda_depth1 = ak.to_backend(depth1, "cuda", highlevel=False)

    assert to_list(ak.count_nonzero(cuda_depth1, -1, highlevel=False)) == to_list(
        ak.count_nonzero(depth1, -1, highlevel=False)
    )
    del cuda_content, cuda_depth1


def test_block_boundary_countnonzero_complex_2():
    rng = np.random.default_rng(seed=42)
    array = rng.integers(2, size=6000)
    complex_array = np.vectorize(complex)(
        array[0 : len(array) : 2], array[1 : len(array) : 2]
    )
    content = ak.contents.NumpyArray(complex_array)
    cuda_content = ak.to_backend(content, "cuda", highlevel=False)
    assert ak.count_nonzero(cuda_content, -1, highlevel=False) == ak.count_nonzero(
        content, -1, highlevel=False
    )

    offsets = ak.index.Index64(np.array([0, 1, 2998, 3000], dtype=np.int64))
    depth1 = ak.contents.ListOffsetArray(offsets, content)
    cuda_depth1 = ak.to_backend(depth1, "cuda", highlevel=False)

    assert to_list(ak.count_nonzero(cuda_depth1, -1, highlevel=False)) == to_list(
        ak.count_nonzero(depth1, -1, highlevel=False)
    )
    del cuda_content, cuda_depth1


def test_block_boundary_argmax_complex1():
    rng = np.random.default_rng(seed=42)
    array = rng.integers(5, size=6000)
    complex_array = np.vectorize(complex)(
        array[0 : len(array) : 2], array[1 : len(array) : 2]
    )
    content = ak.contents.NumpyArray(complex_array)
    cuda_content = ak.to_backend(content, "cuda", highlevel=False)
    assert ak.argmax(cuda_content, -1, highlevel=False) == ak.argmax(
        content, -1, highlevel=False
    )

    offsets = ak.index.Index64(np.array([0, 1, 2998, 3000], dtype=np.int64))
    depth1 = ak.contents.ListOffsetArray(offsets, content)
    cuda_depth1 = ak.to_backend(depth1, "cuda", highlevel=False)
    assert to_list(ak.argmax(cuda_depth1, -1, highlevel=False)) == to_list(
        ak.argmax(depth1, -1, highlevel=False)
    )
    del cuda_content, cuda_depth1


def test_block_boundary_argmax_complex2():
    rng = np.random.default_rng(seed=42)
    array = rng.integers(6000, size=6000)
    complex_array = np.vectorize(complex)(
        array[0 : len(array) : 2], array[1 : len(array) : 2]
    )
    content = ak.contents.NumpyArray(complex_array)
    cuda_content = ak.to_backend(content, "cuda", highlevel=False)
    assert ak.argmax(cuda_content, -1, highlevel=False) == ak.argmax(
        content, -1, highlevel=False
    )

    offsets = ak.index.Index64(np.array([0, 1, 2998, 3000], dtype=np.int64))
    depth1 = ak.contents.ListOffsetArray(offsets, content)
    cuda_depth1 = ak.to_backend(depth1, "cuda", highlevel=False)
    assert to_list(ak.argmax(cuda_depth1, -1, highlevel=False)) == to_list(
        ak.argmax(depth1, -1, highlevel=False)
    )
    del cuda_content, cuda_depth1


def test_block_boundary_argmin_complex1():
    rng = np.random.default_rng(seed=42)
    array = rng.integers(5, size=6000)
    complex_array = np.vectorize(complex)(
        array[0 : len(array) : 2], array[1 : len(array) : 2]
    )
    content = ak.contents.NumpyArray(complex_array)
    cuda_content = ak.to_backend(content, "cuda", highlevel=False)
    assert ak.argmin(cuda_content, -1, highlevel=False) == ak.argmin(
        content, -1, highlevel=False
    )

    offsets = ak.index.Index64(np.array([0, 1, 2998, 3000], dtype=np.int64))
    depth1 = ak.contents.ListOffsetArray(offsets, content)
    cuda_depth1 = ak.to_backend(depth1, "cuda", highlevel=False)
    assert to_list(ak.argmin(cuda_depth1, -1, highlevel=False)) == to_list(
        ak.argmin(depth1, -1, highlevel=False)
    )
    del cuda_content, cuda_depth1


def test_block_boundary_argmin_complex2():
    rng = np.random.default_rng(seed=42)
    array = rng.integers(6000, size=6000)
    complex_array = np.vectorize(complex)(
        array[0 : len(array) : 2], array[1 : len(array) : 2]
    )
    content = ak.contents.NumpyArray(complex_array)
    cuda_content = ak.to_backend(content, "cuda", highlevel=False)
    assert ak.argmin(cuda_content, -1, highlevel=False) == ak.argmin(
        content, -1, highlevel=False
    )

    offsets = ak.index.Index64(np.array([0, 1, 2998, 3000], dtype=np.int64))
    depth1 = ak.contents.ListOffsetArray(offsets, content)
    cuda_depth1 = ak.to_backend(depth1, "cuda", highlevel=False)
    assert to_list(ak.argmin(cuda_depth1, -1, highlevel=False)) == to_list(
        ak.argmin(depth1, -1, highlevel=False)
    )
    del cuda_content, cuda_depth1



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import numpy as np

import awkward as ak


def test_false_parameter():
    dtype = np.dtype("datetime64[15us]")
    form = ak.forms.numpyform.from_dtype(dtype)
    assert form.primitive == "datetime64[15us]"



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import numpy as np
import pytest

import awkward as ak


def test_flat_arrays():
    array = ak.Array(ak.contents.NumpyArray(np.array([1.0, 2.0, 3.0], dtype="f8")))
    form, length, container = ak.to_buffers(array, byteorder="<")
    assert form.is_equal_to(ak.forms.NumpyForm("float64"))
    assert length == 3
    assert container.keys() == {"node0-data"}
    np.testing.assert_array_equal(
        container["node0-data"],
        ak._util.native_to_byteorder(np.array([1.0, 2.0, 3.0], dtype="f8"), "<"),
    )
    reconstructed = ak.from_buffers(form, length, container, byteorder="<")
    assert ak.array_equal(array, reconstructed)

    array = ak.Array(ak.contents.NumpyArray(np.array([1.0, 2.0, 3.0], dtype="f8")))
    form, length, container = ak.to_buffers(array, byteorder=">")
    assert form.is_equal_to(ak.forms.NumpyForm("float64"))
    assert length == 3
    assert container.keys() == {"node0-data"}
    np.testing.assert_array_equal(
        container["node0-data"],
        ak._util.native_to_byteorder(np.array([1.0, 2.0, 3.0], dtype="f8"), ">"),
    )
    reconstructed = ak.from_buffers(form, length, container, byteorder=">")
    assert ak.array_equal(array, reconstructed)


def test_jagged_arrays():
    array = ak.Array(
        ak.contents.ListOffsetArray(
            ak.index.Index(np.array([0, 2, 2, 3], dtype="i8")),
            ak.contents.NumpyArray(np.array([1.0, 2.0, 3.0], dtype="f8")),
        )
    )
    form, length, container = ak.to_buffers(array, byteorder="<")
    assert form.is_equal_to(
        ak.forms.ListOffsetForm("i64", ak.forms.NumpyForm("float64"))
    )
    assert length == 3
    assert container.keys() == {"node0-offsets", "node1-data"}
    np.testing.assert_array_equal(
        container["node0-offsets"],
        ak._util.native_to_byteorder(np.array([0, 2, 2, 3], dtype="i8"), "<"),
    )
    np.testing.assert_array_equal(
        container["node1-data"],
        ak._util.native_to_byteorder(np.array([1.0, 2.0, 3.0], dtype="f8"), "<"),
    )
    reconstructed = ak.from_buffers(form, length, container, byteorder="<")
    assert ak.array_equal(array, reconstructed)

    array = ak.Array(
        ak.contents.ListOffsetArray(
            ak.index.Index(np.array([0, 2, 2, 3], dtype="i8")),
            ak.contents.NumpyArray(np.array([1.0, 2.0, 3.0], dtype="f8")),
        )
    )
    form, length, container = ak.to_buffers(array, byteorder=">")
    assert form.is_equal_to(
        ak.forms.ListOffsetForm("i64", ak.forms.NumpyForm("float64"))
    )
    assert length == 3
    assert container.keys() == {"node0-offsets", "node1-data"}
    np.testing.assert_array_equal(
        container["node0-offsets"],
        ak._util.native_to_byteorder(np.array([0, 2, 2, 3], dtype="i8"), ">"),
    )
    np.testing.assert_array_equal(
        container["node1-data"],
        ak._util.native_to_byteorder(np.array([1.0, 2.0, 3.0], dtype="f8"), ">"),
    )
    reconstructed = ak.from_buffers(form, length, container, byteorder=">")
    assert ak.array_equal(array, reconstructed)


def test_flat_bytes():
    form = ak.forms.NumpyForm("float64", form_key="node0")
    length = 3
    container = {"node0-data": np.array([1.0, 2.0, 3.0], dtype="<f8").tobytes()}
    array = ak.from_buffers(form, length, container, byteorder="<")
    assert ak.array_equal(
        array, ak.Array(ak.contents.NumpyArray(np.array([1.0, 2.0, 3.0], dtype="f8")))
    )

    form = ak.forms.NumpyForm("float64", form_key="node0")
    length = 3
    container = {"node0-data": np.array([1.0, 2.0, 3.0], dtype="<f8").tobytes()}
    array = ak.from_buffers(form, length, container, byteorder=">")
    assert ak.array_equal(
        array,
        ak.Array(
            ak.contents.NumpyArray(np.array([1.0, 2.0, 3.0], dtype="f8").byteswap())
        ),
    )

    form = ak.forms.NumpyForm("float64", form_key="node0")
    length = 3
    container = {"node0-data": np.array([1.0, 2.0, 3.0], dtype=">f8").tobytes()}
    array = ak.from_buffers(form, length, container, byteorder="<")
    assert ak.array_equal(
        array,
        ak.Array(
            ak.contents.NumpyArray(np.array([1.0, 2.0, 3.0], dtype="f8").byteswap())
        ),
    )

    form = ak.forms.NumpyForm("float64", form_key="node0")
    length = 3
    container = {"node0-data": np.array([1.0, 2.0, 3.0], dtype=">f8").tobytes()}
    array = ak.from_buffers(form, length, container, byteorder=">")
    assert ak.array_equal(
        array, ak.Array(ak.contents.NumpyArray(np.array([1.0, 2.0, 3.0], dtype="f8")))
    )


def test_jagged_bytes():
    form = ak.forms.ListOffsetForm(
        "i64", ak.forms.NumpyForm("float64", form_key="node1"), form_key="node0"
    )
    length = 3
    container = {
        "node0-offsets": np.array([0, 2, 2, 3], dtype="<i8").tobytes(),
        "node1-data": np.array([1.0, 2.0, 3.0], dtype="<f8").tobytes(),
    }
    array = ak.from_buffers(form, length, container, byteorder="<")
    assert ak.array_equal(
        array,
        ak.Array(
            ak.contents.ListOffsetArray(
                ak.index.Index(np.array([0, 2, 2, 3], dtype="i8")),
                ak.contents.NumpyArray(np.array([1.0, 2.0, 3.0], dtype="f8")),
            )
        ),
    )

    form = ak.forms.ListOffsetForm(
        "i64", ak.forms.NumpyForm("float64", form_key="node1"), form_key="node0"
    )
    length = 3
    container = {
        "node0-offsets": np.array([0, 2, 2, 3], dtype="<i8").tobytes(),
        "node1-data": np.array([1.0, 2.0, 3.0], dtype="<f8").tobytes(),
    }
    with pytest.raises((ValueError, OverflowError)):
        array = ak.from_buffers(form, length, container, byteorder=">")

    form = ak.forms.ListOffsetForm(
        "i64", ak.forms.NumpyForm("float64", form_key="node1"), form_key="node0"
    )
    length = 3
    container = {
        "node0-offsets": np.array([0, 2, 2, 3], dtype=">i8").tobytes(),
        "node1-data": np.array([1.0, 2.0, 3.0], dtype=">f8").tobytes(),
    }
    with pytest.raises((ValueError, OverflowError)):
        array = ak.from_buffers(form, length, container, byteorder="<")

    form = ak.forms.ListOffsetForm(
        "i64", ak.forms.NumpyForm("float64", form_key="node1"), form_key="node0"
    )
    length = 3
    container = {
        "node0-offsets": np.array([0, 2, 2, 3], dtype=">i8").tobytes(),
        "node1-data": np.array([1.0, 2.0, 3.0], dtype=">f8").tobytes(),
    }
    array = ak.from_buffers(form, length, container, byteorder=">")
    assert ak.array_equal(
        array,
        ak.Array(
            ak.contents.ListOffsetArray(
                ak.index.Index(np.array([0, 2, 2, 3], dtype="i8")),
                ak.contents.NumpyArray(np.array([1.0, 2.0, 3.0], dtype="f8")),
            )
        ),
    )



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import pytest

import awkward as ak

nb = pytest.importorskip("numba")
nb_cuda = pytest.importorskip("numba.cuda")

from numba import types  # noqa: E402

nb_cuda.config.CUDA_LOW_OCCUPANCY_WARNINGS = False
nb_cuda.config.CUDA_WARN_ON_IMPLICIT_COPY = False


try:
    ak.numba.register_and_check()
except ImportError:
    pytest.skip(reason="too old Numba version", allow_module_level=True)


def test_array_typed():
    # create an ak.Array with a cuda backend:
    gpu_arr_type = ak.Array([[[0, 1], [2, 3]], [[4, 5]]], backend="cuda").numba_type

    @nb.cuda.jit(types.void(gpu_arr_type), extensions=[ak.numba.cuda])
    def cuda_kernel(arr):
        return None

    array = ak.Array([[[0, 1], [2, 3]], [[4, 5]]], backend="cuda")
    cuda_kernel[1024, 1024](array)

    other_array = ak.Array(
        [[1.1, 2.2, 3.3], [None, 4.4], None, [None, 5.5]], backend="cuda"
    )

    with pytest.raises(nb.core.errors.NumbaTypeError):
        cuda_kernel[1024, 1024](other_array)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import awkward as ak
from awkward._dispatch import high_level_function
from awkward._layout import HighLevelContext, maybe_posaxis
from awkward._namedaxis import (
    _get_named_axis,
    _named_axis_to_positional_axis,
)
from awkward._nplikes.numpy_like import NumpyMetadata
from awkward._regularize import regularize_axis
from awkward.errors import AxisError

__all__ = ("drop_none",)

np = NumpyMetadata.instance()


@high_level_function()
def drop_none(array, axis=None, highlevel=True, behavior=None, attrs=None):
    """
    Args:
        array: Data in which to remove Nones.
        axis (None or int or str): If None, the operation drops Nones at all levels
            of nesting, returning an array of the same dimension, but without Nones.
            If an int, the depth at which to drop Nones.
            The outermost dimension is `0`, followed by `1`, etc., and negative
            values count backward from the innermost: `-1` is the innermost
            dimension, `-2` is the next level up, etc.
            If a str, it is interpreted as the name of the axis which maps to
            an int if named axes are present.
            Named axes are attached to an array using #ak.with_named_axis and
            removed with #ak.without_named_axis; also see the
            [Named axes user guide](../../user-guide/how-to-array-properties-named-axis.html).
        highlevel (bool): If True, return an #ak.Array; otherwise, return
            a low-level #ak.contents.Content subclass.
        behavior (None or dict): Custom #ak.behavior for the output array, if
            high-level.
        attrs (None or dict): Custom attributes for the output array, if
            high-level.

    Removes missing values (None) from a given array.

    For example, in the following `array`,

        >>> array = ak.Array([[[0]], [[None]], [[1], None], [[2, None]]])

    The None value will be removed, resulting in

        >>> ak.drop_none(array)
        <Array [[[0]], [[]], [[1]], [[2]]] type='4 * var * var * int64'>

    The default axis is None, however an axis can be specified:

        >>> ak.drop_none(array, axis=1)
        <Array [[[0]], [[None]], [[1]], [[2, None]]] type='4 * var * var * ?int64'>

    """
    # Dispatch
    yield (array,)

    # Implementation
    return _impl(array, axis, highlevel, behavior, attrs)


def _drop_none_if_list(layout):
    if layout.is_list:
        # only drop nones at list level in the recursion; this way ListArray -> ListOffsetArray with unprojected optiontype -> avoid offset mismatch
        return layout.drop_none()
    else:
        return layout


def _impl(array, axis, highlevel, behavior, attrs):
    with HighLevelContext(behavior=behavior, attrs=attrs) as ctx:
        layout = ctx.unwrap(array, allow_record=False, primitive_policy="error")

    # Handle named axis
    named_axis = _get_named_axis(ctx)
    # Step 1: Normalize named axis to positional axis
    axis = _named_axis_to_positional_axis(named_axis, axis)

    axis = regularize_axis(axis, none_allowed=True)

    if axis is None:
        # if the outer layout is_option, drop_nones without affecting offsets
        if layout.is_option:
            layout = layout.drop_none()

        def action(layout, continuation, **kwargs):
            return _drop_none_if_list(continuation())

    else:
        max_axis = layout.branch_depth[1] - 1
        if axis > max_axis:
            raise AxisError(f"axis={axis} exceeds the depth ({max_axis}) of this array")

        def recompute_offsets(layout, depth, **kwargs):
            posaxis = maybe_posaxis(layout, axis, depth)
            if (posaxis == 0 and posaxis == depth - 1) or (
                posaxis == depth and layout.is_list
            ):
                none_indexes = options["none_indexes"].pop(0)
                out = layout._rebuild_without_nones(none_indexes, layout.content)
                return out

        def action(layout, depth, **kwargs):
            if layout.is_record:
                posaxises = {maybe_posaxis(x, axis, depth) for x in layout.contents}
                if len(posaxises) > 1 and any(x < depth for x in posaxises):
                    raise AxisError(
                        f"axis={axis} implies different levels in records that might require part of a record to be dropped, which is impossible"
                    )
            posaxis = maybe_posaxis(layout, axis, depth)
            if posaxis == 0:
                if not layout.is_option:
                    return layout
                else:
                    return layout.drop_none()
            if posaxis == depth - 1 and layout.is_option:
                _, _, none_indexes = layout._nextcarry_outindex()
                options["none_indexes"].append(none_indexes)
                return layout.drop_none()
            if posaxis == depth - 1 and layout.is_list and layout.content.is_option:
                return layout.drop_none()

    options = {"none_indexes": []}
    out = ak._do.recursively_apply(layout, action, depth_context=options)

    if len(options["none_indexes"]) > 0:
        out = ak._do.recursively_apply(out, recompute_offsets, depth_context=options)

    return ctx.wrap(
        out,
        highlevel=highlevel,
    )



from __future__ import annotations

import pytest

import awkward as ak

pa = pytest.importorskip("pyarrow")


def test_tz_is_dropped():
    data = pa.Table.from_arrays(
        [
            pa.array(
                [
                    1,
                    2,
                    3,
                ],
                type=pa.timestamp("ns", tz="UTC"),
            )
        ],
        names=["a"],
    )
    ak.from_arrow(data)
    ak.from_arrow_schema(data.schema)
